#------------------------------------------------------------------
# $Copyright: Copyright (c) 2011 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/SAPNW04/SAPNW04.pm,v 1.77 2010/11/01 06:41:21 rkumar Exp $
#
# $Author: rkumar $
# $Date: 2010/11/01 06:41:21 $
# $Revision: 1.77 $
#
# Provides:                            SAPNW04
# Author:                              vchappav
# Creation Date:                       16 Feb, 2006
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SAPNW04;

my ( $CLUS_HOME, $sPrefix ) = ();

#---------------------
# Setup VCS/VCSOne home..
#---------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.0/VCSOne..
  $CLUS_HOME = $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS
  $CLUS_HOME = $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vcs(|one))$/;

if ( $sPrefix =~ /vcsone/ ) {
   $sPrefix = "ha";
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = "$CLUS_HOME/bin/hares";
my $sCmdHasys  = "$CLUS_HOME/bin/hasys";
my $sCmdHagrp  = "$CLUS_HOME/bin/hagrp";
my $sCmdHatype = "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( uname );
use POSIX qw ( strftime );
use File::Path qw ( mkpath );

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SAPNW04.pm,v 1.77 2010/11/01 06:41:21 rkumar Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#-----------------
# Define globals..
#-----------------
my ( $iSapStartSrv, $iCCMSX, $iCCMSR, $bSapStart, @liRunningProcs ) = (); # Global..

#-----------------------------
# Define SAPNW04 subroutines..
#-----------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnqSrvResName,
                              IN String EnvFile,
                              IN String InstName,
                              IN String InstType,
                              IN String MonitorProgram,
                              IN String ProcMon,
                              IN String SAPAdmin,
                              IN String SAPMonHome,
                              IN String SAPSID,
                              IN Integer SecondLevelMonitor,
                              IN String StartProfile );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localEnqSrvResName            Name of the standalone ENQUEUE server VCS resource
 $localEnvFile                  Full path to the environment file for SAP
 $localInstName                 String identifier that uniquely identifies a SAP server instance
 $localInstType                 String identifier that classifies the SAP server instance type
 $localMonitorProgram           Full path to external monitor program
 $localProcMon                  List of SAP processes to monitor
 $localSAPAdmin                 UNIX user dedicated for this instance of SAP
 $localSAPMonHome               Directory defining the location of the SAP monitoring utilities
 $localSAPSID                   SAP system name defined during installation
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localStartProfile             Specifies the full path to the start-profile for this instance
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnqSrvResName,
       $localEnvFile,
       $localInstName,
       $localInstType,
       $localMonitorProgram,
       $localProcMon,
       $localSAPAdmin,
       $localSAPMonHome,
       $localSAPSID,
       $localSecondLevelMonitor,
       $localStartProfile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnqSrvResName            [%s]
  EnvFile                  [%s]
  InstName                 [%s]
  InstType                 [%s]
  MonitorProgram           [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPMonHome               [%s]
  SAPSID                   [%s]
  SecondLevelMonitor       [%s]
  StartProfile             [%s]",

  $sMyName, $sEntryPointName,  $localEnqSrvResName,  $localEnvFile,
  $localInstName,       $localInstType,       $localMonitorProgram,
  $localProcMon,  $localSAPAdmin,  $localSAPMonHome,  $localSAPSID,
  $localSecondLevelMonitor, $localStartProfile );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName   EnqSrvResName    EnvFile    InstName    InstType
       MonitorProgram     ProcMon     SAPAdmin     SAPMonHome     SAPSID
       SecondLevelMonitor StartProfile );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EnvFile InstName InstType ProcMon SAPAdmin SAPSID StartProfile );
  my @lsOfflineArgs = qw ( InstName InstType SAPAdmin SAPSID StartProfile );
  my @lsMonitorArgs = qw ( InstName InstType ProcMon SAPAdmin SAPSID StartProfile );
  my @lsCleanArgs   = qw ( InstName InstType SAPAdmin SAPSID StartProfile );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{$hlsNeeded { $sEntryPointName }};

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s]. Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  VCSagentFW::RunSecondLevelMonitor ( 1 ) if ( $sEntryPointName ne 'monitor' );

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------
  # Individual validation starts..
  #-------------------------------

  #--------------------------------------------------
  # StartProfile must terminate in a valid hostname..
  #--------------------------------------------------
  ( my $sHost = $localStartProfile ) =~ s#.*/.*?_[^_]+_##;

  if ( $sEntryPointName eq 'online' ) {
    if ( ! inet_aton ( $sHost ) ) {
      &logError ( 20003,
      "<%s> StartProfile [%s] not terminated with valid hostname [%s]",
      $sMyName, $localStartProfile, $sHost );
      $iReturnCode = 0; # Failure..
    }
  }

  #-----------------------------------
  # InstName. Last two chars 00 - 99..
  #-----------------------------------
  unless ( $localInstName =~ /(\d{2})$/o ) {          # && $1 <= 99 ) 
    &logError ( 20004,
    "<%s> Last two characters of InstName [%s] not between 00-99",
    $sMyName, $localInstName );
    $iReturnCode = 0; # Failure..
  }

  #---------------------------
  # Check if SAPAdmin exists..
  #---------------------------
  if ( $sEntryPointName =~ /^(online|offline|monitor)$/o ) {
    if ( ! defined getpwnam $localSAPAdmin ) {
      &logError ( 20005,
      "<%s> SAPAdmin [%s] not a valid UNIX login",
      $sMyName, $localSAPAdmin );
      $iReturnCode = 0; # Failure..
    }
  }

  #-----------------------------------------
  # SAPSID.. must be three characters long..
  #-----------------------------------------
  if ( $localSAPSID !~ /^\S{3}$/o ) {
    &logError ( 20006,
    "<%s> SAPSID [%s] has to be exactly 3 non-white characters long",
    $sMyName, $localSAPSID );
    $iReturnCode = 0; # Failure..
  }

  if ( $localSAPSID !~ /^[a-z]/io ) {
    &logError ( 20007,
    "<%s> SAPSID [%s] has to begin with an alphabet",
    $sMyName, $localSAPSID );
    $iReturnCode = 0; # Failure..
  }

  #------------------
  # Valid InstTypes..
  #------------------
  my @lsIntTypes = qw ( central dialog enqueue aenqueue jenqueue enqrep aenqrep jenqrep );

  unless ( grep { /^$localInstType$/io } @lsIntTypes ) {
    &logError ( 20008,
    "<%s> InstType [%s] invalid. Has to be one of [%s]",
    $sMyName, $localInstType, join ( ' ', @lsIntTypes ) );
    $iReturnCode = 0; # Failure..
  }

  if ( $sEntryPointName =~ /^(online|monitor)$/o ) {
    #---------------------------------------------
    # Validate ProcMon after removing duplicates..
    #--------------------------------------------
    my $rslocalProcMon = \$_[6]; # Get reference to what's passed in.. 
    my %hsProcMon = map { $_ => 1 } $localProcMon =~ /([^\s]+)/g;
    my @lsProcMon = keys %hsProcMon;
    my ( @lsBadProcs, %hsAllowed ) = ();

    if ( $localInstType =~ /^(dialog|central)$/io ) {
      if ( ! grep { /^(dw|jc)$/o } @lsProcMon ) { # Must have..
        &logError ( 20009, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least one of [dw jc] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }

      if ( $localInstType =~ /^central$/io ) {
        %hsAllowed = map { $_ => 1 } qw ( dw ms co se jc ig gwrd icman ); # Restricted set..
        if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
          &logError ( 20010, "<%s> Incorrect value(s) [%s] in ProcMon found. 
          Must be one of [%s] for InstType [%s]", $sMyName,
          join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
          $localInstType );

          $iReturnCode = 0; # Failure..
        }
      }
      elsif ( $localInstType =~ /^dialog$/io ) {
        %hsAllowed = map { $_ => 1 } qw ( dw se jc ig gwrd icman ); # Restricted set..
        if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
          &logError ( 20011, "<%s> Incorrect value(s) [%s] in ProcMon found. 
          Must be one of [%s] for InstType [%s]", $sMyName,
          join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
          $localInstType );

          $iReturnCode = 0; # Failure..
        }
      }
    }
    elsif ( $localInstType =~ /^(|a|j)enqueue$/io ) {
      if ( ! grep { /^en$/o } @lsProcMon ) {
        &logError ( 20012, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least [en] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }

      %hsAllowed = map { $_ => 1 } qw ( co en ms se ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20013, "<%s> Incorrect value(s) [%s] in ProcMon found. 
        Must be one of [%s] for InstType [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
        $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif ( $localInstType =~ /^(|a|j)enqrep$/io  ) {
      if ( ! grep { /^e(|n)r$/o } @lsProcMon ) {
        &logError ( 20014, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least [er enr] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }

      %hsAllowed = map { $_ => 1 } qw ( er enr ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20015, "<%s> Incorrect value(s) [%s] in ProcMon found. 
        Must be one of [%s] for InstType [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
        $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }

    #--------------------
    # Re-create procmon..
    #--------------------
    $$rslocalProcMon = join ( ' ', @lsProcMon );
  }

  #--------------
  # Online only..
  #--------------
  if ( $sEntryPointName eq 'online' ) {
    #----------------------------------
    # Check if EnqSrvResName is valid..
    #----------------------------------
    if ( $localEnqSrvResName =~ m/\S+/ && $localInstType =~ /^(|a|j)enqrep$/io  &&
         ! SAPNW04::ChkResName ( $localEnqSrvResName,
                                 $localSAPSID,
                                 $localInstType ) ) {
      &logError ( 20016,
      "<%s> EnqSrvResName [%s] invalid resource of Type [SAPNW04]",
      $sMyName, $localEnqSrvResName );
      $iReturnCode = 0; # Failure..
    }

    #-----------------------
    # EnvFile has to exist..
    #-----------------------
    if ( ! -e $localEnvFile ) {
      &logError ( 20017,
      "<%s> EnvFile [%s] does not exist", $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }
   
    #----------------------------
    # StartProfile has to exist..
    #----------------------------
    if ( ! -r $localStartProfile ) {
      &logError ( 20018,
      "<%s> StartProfile [%s] does not exist",
      $sMyName, $localStartProfile );
      $iReturnCode = 0; # Failure..
    }

    #-------------------
    # Validate SAPHome..
    #-------------------
    if (  $localSecondLevelMonitor > 0 ) {
      if ( ! ( $localSAPMonHome =~ /\S+/ ) || 
           ! ( -d $localSAPMonHome ) ) {
        &logError ( 20019,
        "<%s> SAPMonHome [%s] not a valid directory",
        $sMyName, $localSAPMonHome );
        $iReturnCode = 0; # Failure..
      }
    }

    #--------------------------------------
    # MonitorProgram has to be executable..
    #--------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {                        # And not executable..
      &logError ( 20020,
      "<%s> MonitorProgram [%s] not executable", $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20021,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String EnvFile,
                         IN String EnqSrvResName,
                         IN String InstName,
                         IN String InstType,
                         IN String ProcMon,
                         IN String SAPAdmin,
                         IN String SAPMonHome,
                         IN String SAPSID,
                         IN String StartProfile,
                         IN Integer SecondLevelMonitor );

 Description : Start logic needed for online entry pt

 Parameters :

 $sEnvFile            Full path to the environment file for SAP
 $sEnqSrvResName      Name of the standalone ENQUEUE server VCS resource
 $sInstName           String identifier that uniquely identifies a SAP server instance
 $sInstType           String identifier that classifies the SAP server instance type
 $sProcMon            List of SAP processes to monitor
 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sSAPMonHome         Directory defining the location of the SAP monitoring utilities
 $sSAPSID             SAP system name defined during installation
 $sStartProfile       Specifies the full path to the start-profile for this instance
 $iSecondLevelMonitor Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sEnqSrvResName,
       $sInstName,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sSAPMonHome,
       $sSAPSID,
       $sStartProfile,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20022, "Subroutine <%s> called with:
  EnvFile                  [%s]
  EnqSrvResName            [%s]
  InstName                 [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPMonHome               [%s]
  SAPSID                   [%s]
  StartProfile             [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,
  $sEnvFile,  $sEnqSrvResName,  $sInstName,  $sInstType, $sProcMon,
  $sSAPAdmin, $sSAPMonHome, $sSAPSID, $sStartProfile,
  $iSecondLevelMonitor );
  
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTillWhen, $iTimeout, $sKillFile, $sCollLockFile,
       $sSendLockFile, @lsFiles, @liPids ) = ();

  #-------------------
  # Initiate SAP log..
  #-------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sMessage = "Starting up [$sResourceName]";

  if ( SAPNW04::UpdateSAPLog ( $sInstName,
                               $sInstType,
                               $sSAPAdmin,
                               $sSAPSID,
                               $sStartProfile,
                               $sMessage ) ) {
    &logTrace ( 20023,
    "<%s> Updated log successfully", $sMyName );
  }
  else {
    &logError ( 20024,
    "<%s> Could not log successfully", $sMyName );
  }

  #----------------------------
  # Remove existing processes..
  #----------------------------
  if ( SAPNW04::KillSAPProcs ( $sInstName,
                               $sProcMon,
                               $sSAPAdmin,
                               $sSAPSID,
                               $sStartProfile ) ) {
    &logTrace ( 20025,
    "<%s> Confirmed no persistent processes exist before startup",
    $sMyName );
  }
  else {
    &logError ( 20026,
    "<%s> Persistent processes exist for SAP instance [%s]",
    $sMyName, $sInstName );
    return 0; # Failure..
  }

  #--------------------
  # Get time bearings..
  #--------------------
  $iTillWhen = $^T + SAPNW04::GetEngineTimeout ( 1 ); # With margin..
  $iTimeout = $iTillWhen - time;

  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( $sInstType =~ /^(central|dialog|enqrep|aenqrep|jenqrep)$/io ) {
    #----------------------
    # Clean IPC resources..
    #----------------------
    if ( SAPNW04::CleanIPC ( $sSAPAdmin,
                             $sEnvFile,
                             $sSAPSID,
                             $sInstName,
                             $iTimeout ) ) {
      &logTrace ( 20027,
      "<%s> Executed cleanipc successfully", $sMyName );
    }
    else {
      $iReturnCode = 0; # Failure..
      &logError ( 20028,
      "<%s> Failed to execute cleanipc. Returning with [%s]",
      $sMyName, $iReturnCode );
    }
  }
  else {
    &logTrace ( 20029,
    "<%s> Skipping cleanipc for InstType [%s].
    Possibility of destroying active lock table", $sMyName, $sInstType );
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #------------------
  # Remove kill.sap..
  #------------------
  $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

  if ( -e $sKillFile ) {
    &logTrace ( 20030,
    "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
    push ( @lsFiles, $sKillFile );
  }

  #-----------------------------
  # Remove collector lock file..
  #-----------------------------
  $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

  if ( -e $sCollLockFile ) {
    &logTrace ( 20031,
    "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
    push ( @lsFiles, $sCollLockFile );
  }

  #--------------------------
  # Remove sender lock file..
  #--------------------------
  $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

  if ( -e $sSendLockFile ) {
    &logTrace ( 20032,
    "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
    push ( @lsFiles, $sSendLockFile );
  }

  #-------------------
  # Remove the files..
  #-------------------
  if ( scalar @lsFiles &&
       ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
    &logError ( 20033, "<%s> Failed to remove stale file(s) [%s]",
    $sMyName, join ( ' ', @lsFiles ) );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #------------------------------------
  # Attempt to start SAP OS collector..
  #------------------------------------
  my $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/saposcol";
  my $sCmdArgs = "";
  my ( $iExitCode, @lsCmdOut ) = ();

  #-----------------------
  # Execute if validated..
  #-----------------------
  if ( ! -e $sCmd ) {
    &logError ( 20034,
    "<%s> SAP OS Collector binary [%s] does not exist",
    $sMyName, $sCmd );
  }
  elsif ( !-x $sCmd ) {
    &logError ( 20035,
    "<%s> SAP OS Collector binary [%s] is not executable",
    $sMyName, $sCmd );
  }
  else {
    $iTimeout = $iTillWhen - time;      # Get time bearings..
    if ( SAPNW04::ExecLocalCmd ( $sSAPAdmin, # User context..
                                 $sEnvFile,  # envfile..
                                 $sCmd,
                                 $sCmdArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdOut ) ) {
      &logTrace ( 20036,
      "<%s> Started SAP OS Collector [%s] successfully with exit code [%s]",
      $sMyName, $sCmd, $iExitCode );
    }
    elsif ( ! defined $iExitCode ) { # Time-out..
      &logError ( 20037,
      "<%s> Cmd [%s] timed out in [%s] sec", $sMyName, $sCmd, $iTimeout );
      return 0; # Failure. Get out..
    }
    elsif ( 255 == $iExitCode ) {  # Up already..
      &logInfo ( 20038,
      "<%s> SAP OS Collector [%s] already up", $sMyName, $sCmd );
    }
    else { # Error..
      &logError ( 20039,
      "<%s> Cmd [%s] failed with exit code [%s]. Cmd output [%s]",
      $sMyName, $sCmd, $iExitCode, join ( chr 012, @lsCmdOut ) );
    }
  }

  #---------------------------------------------
  # Attempt to start SAP services for SAPNW04s..
  #---------------------------------------------
  $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/sapstartsrv";
  $sCmdArgs = "pf=$sStartProfile -D -u $sSAPAdmin";
  ( $iExitCode, @lsCmdOut ) = (); # Reset..

  #-----------------------
  # Execute if validated..
  #-----------------------
  if ( ! -e $sCmd ) {
    &logTrace ( 20040,
    "<%s> SAP services binary [%s] does not exist",
    $sMyName, $sCmd );
  }
  elsif ( !-x $sCmd ) {
    &logTrace ( 20041,
    "<%s> SAP services binary [%s] is not executable",
    $sMyName, $sCmd );
  }
  else {
    $iTimeout = $iTillWhen - time;       # Get time bearings..
    if ( SAPNW04::ExecLocalCmd ( $sSAPAdmin, # User context..
                                 $sEnvFile,  # envfile..
                                 $sCmd,
                                 $sCmdArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdOut ) ) {
      &logTrace ( 20042,
      "<%s> Started SAP services [%s] successfully with exit code [%s]",
      $sMyName, $sCmd, $iExitCode );
    }
    elsif ( ! defined $iExitCode ) { # Time-out..
      &logError ( 20043,
      "<%s> Cmd [%s] timed out in [%s] sec", $sMyName, $sCmd, $iTimeout );
      return 0; # Failure. Get out..
    }
    else { # Error..
      &logError ( 20044,
      "<%s> Cmd [%s] failed with exit code [%s]. Cmd output [%s]",
      $sMyName, $sCmd, $iExitCode, join ( chr 012, @lsCmdOut ) );
    }
  }

  #-------------------------
  # Start the SAP instance..
  #-------------------------
  my %hsExitCode = ( 0 => 'Success',
                     1 => 'OS configuration error',
                     2 => 'Called with wrong Parameters',
                     3 => 'User not authorized to execute script',
                     4 => 'Environment not set',
                     5 => 'Database startup failed',
                     6 => 'Database on remote DB-Server not running',
                     7 => 'Startup of Instance failed',
                     8 => 'Start profile not found',
                     9 => 'User not authorized to start database' );

  my $sLogFile = "/usr/sap/$sSAPSID/$sInstName/log/" .
                 "${sSAPSID}_${sInstName}.log";
  my $iSleepTime = 10; # Secs..

  #-------------------------------------------
  # Get re-direct string based on user shell..
  #-------------------------------------------
  my $sRdStr = ( getpwnam $sSAPAdmin ) [-1] =~ /csh$/o ?">>& $sLogFile":
                                                        "1>> $sLogFile 2>&1";
  $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/sapstart";
  $sCmdArgs = "pf=$sStartProfile $sRdStr";

  ( $iExitCode, @lsCmdOut ) = (); # Reset..

  #-----------------------
  # Execute if validated..
  #-----------------------
  if ( ! -e $sCmd ) {
    &logError ( 20045,
    "<%s> SAP start command [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( !-x $sCmd ) {
    &logError ( 20046,
    "<%s> SAP start command [%s] is not executable",
    $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20047,
    "<%s> In-sufficient time [%s] to execute Cmd [%s] with Args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = 0; # Failure..
  }
  elsif ( SAPNW04::ExecLocalCmd ( $sSAPAdmin, # User context..
                                  $sEnvFile,  # Use envfile..
                                  $sCmd,
                                  $sCmdArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdOut ) ) {
    &logTrace ( 20048,
    "<%s> Started SAP instance [%s] successfully with exit code [%s]",
    $sMyName, $sInstName, $iExitCode );

    %hsExitCode = (); # Free..

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
      &logTrace ( 20049,
      "<%s> Confirming start-up with [%s] sec in hand",
      $sMyName, $iTimeout );

      #------------------
      # Check on status..
      #------------------
      if ( VCSDef_MonitorExit::ONLINE ==
           SAPNW04::GetProcState ( $sInstName,
                                   $sProcMon,
                                   $sSAPAdmin,
                                   $sSAPSID,
                                   $sStartProfile,
                                   \@liPids,        # Online after..
                                   1 )         && ( # Running FFLM..
           $iSecondLevelMonitor ?                   # If enabled..
           VCSDef_MonitorExit::ONLINE ==            # Online after..
           SAPNW04::SecondLevelMonitor ( $sEnvFile, # Running SLM..
                                         $sInstName,
                                         $sInstType,
                                         $sSAPAdmin,
                                         $sProcMon,
                                         $sSAPMonHome,
                                         $sSAPSID,
                                         $sStartProfile,
                                         $iTillWhen ):1 ) ) {
        &logInfo ( 20050,
        "<%s> Confirmed start-up of SAP Instance with PIDs [%s]",
        $sMyName, join ( ' ', @liPids ) );

        $iReturnCode = 1; # Success..
        last; # Get out..
      }
      else {
        &logTrace ( 20051,
        "<%s> Could not confirm start-up. Sleeping for [%s] sec",
        $sMyName, $iSleepTime );

        $iReturnCode = 0;  # Failure..
        sleep $iSleepTime; # Sec..
      }
    }
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    &logError ( 20052,
    "<%s> Cmd [%s] with Args [%s] timed out in [%s] sec",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout );
    $iReturnCode = 0; # Failure..
  }
  else { # Error..
    $hsExitCode { $iExitCode } ||= 'undef'; # Prevent undefs..
    &logError ( 20053,
    "<%s> Cmd [%s] failed with exit code [%s]:[%s]. Cmd output [%s]",
    $sMyName, $sCmd, $iExitCode, $hsExitCode { $iExitCode },
    join ( chr 012, @lsCmdOut ) );
    $iReturnCode = 0; # Failure..
  }

  %hsExitCode = (); # Free..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String EnvFile,
                        IN String InstName,
                        IN String InstType,
                        IN String ProcMon,
                        IN String SAPAdmin,
                        IN String SAPSID,
                        IN String StartProfile );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sEnvFile      Full path to the environment file for SAP
 $sInstName     String identifier that uniquely identifies a SAP server instance
 $sInstType     String identifier that classifies the SAP server instance type
 $sProcMon      List of SAP processes to monitor
 $sSAPAdmin     UNIX user dedicated for this instance of SAP
 $sSAPSID       SAP system name defined during installation
 $sStartProfile Specifies the full path to the start-profile for this instance

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstName,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sSAPSID,
       $sStartProfile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20054, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstName                 [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  StartProfile             [%s]",

  $sMyName, $sEnvFile, $sInstName, $sInstType, $sProcMon,
  $sSAPAdmin, $sSAPSID, $sStartProfile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iSleepTime, $iTillWhen, $iTimeout, $sKillFile,
       $sCollLockFile, $sSendLockFile, @lsFiles, @liPids ) = ();

  #--------------------
  # Get time bearings..
  #--------------------
  $iTillWhen = $^T + SAPNW04::GetEngineTimeout ( 1 ); # With margin..
  $iSleepTime = 10; # Secs..

  @liPids = $bSapStart ? shift @liRunningProcs : @liRunningProcs; # Global..

  if( $bSapStart && defined $iCCMSX ) {
    push ( @liPids, $iCCMSX );
  }

  if( $bSapStart && defined $iCCMSR ) {
    push ( @liPids, $iCCMSR );
  }

  #-----------------------------------------------
  # Attempt to request sapstart to stop instance..
  #-----------------------------------------------
  if ( scalar @liPids ) {
    &logInfo ( 20055, "<%s> Sending SIGINT to PIDs [%s]",
    $sMyName, join ( ' ', @liPids ) );
  
    #-----------------------------------------
    # Give SAP chance to shutdown gracefully..
    #-----------------------------------------
    if ( kill ( 2, @liPids ) == scalar @liPids ) {
      #--------------------
      # Delay entry point..
      #--------------------
      while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
        #-----------------------
        # Wait before retrying..
        #-----------------------
        sleep $iSleepTime; # Secs..

        SAPNW04::GetProcState ( $sInstName,
                                $sProcMon,
                                $sSAPAdmin,
                                $sSAPSID,
                                $sStartProfile,
                                \@liPids,
                                0 );

        #------------------------------
        # Check for pending processes..
        #------------------------------
        if ( scalar @liPids ) {
          #-----------------------------------------------
          # Check if the last process is not sapstartsrv..
          #-----------------------------------------------
          if ( 1 == scalar @liPids   && # Last remaining..
                  defined $iSapStartSrv && #
                  $iSapStartSrv == $liPids [0] ) {
            last;
          }

          &logTrace ( 20056,
          "<%s> Could not confirm shutdown. Retrying in [%s] sec.
          Delaying kill with [%s] sec in hand",
          $sMyName, $iSleepTime, $iTimeout );
        }
        else {
          &logInfo ( 20057,
          "<%s> Confirmed shutdown of SAP Instance [%s]",
          $sMyName, $sInstName );
          last; # Get out..
        }
      }
    }
    else {
      &logWarn ( 20058,
      "<%s> Failed to send SIGINT to PID [%s]",
      $sMyName, join ( ' ', @liPids ) );
    }
  }

  #----------------------------
  # Remove existing processes..
  #----------------------------
  if ( SAPNW04::KillSAPProcs ( $sInstName,
                               $sProcMon,
                               $sSAPAdmin,
                               $sSAPSID,
                               $sStartProfile ) ) {
    &logTrace ( 20059,
    "<%s> Confirmed no persistent processes exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20060,
    "<%s> Persistent processes exist for SAP instance [%s]",
    $sMyName, $sInstName );
    return 0; # Failure..
  }

  #------------------
  # Remove kill.sap..
  #------------------
  $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

  if ( -e $sKillFile ) {
    &logTrace ( 20061,
    "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
    push ( @lsFiles, $sKillFile );
  }

  #-----------------------------
  # Remove collector lock file..
  #-----------------------------
  $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

  if ( -e $sCollLockFile ) {
    &logTrace ( 20062,
    "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
    push ( @lsFiles, $sCollLockFile );
  }

  #--------------------------
  # Remove sender lock file..
  #--------------------------
  $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

  if ( -e $sSendLockFile ) {
    &logTrace ( 20063,
    "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
    push ( @lsFiles, $sSendLockFile );
  }

  #-------------------
  # Remove the files..
  #-------------------
  if ( scalar @lsFiles &&
       ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
    &logError ( 20064, "<%s> Failed to remove stale file(s) [%s]",
    $sMyName, join ( ' ', @lsFiles ) );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( $sInstType =~ /^(central|dialog|enqrep|aenqrep|jenqrep)$/io ) {
    #----------------------
    # Clean IPC resources..
    #----------------------
    $iTimeout = $iTillWhen - ( time + 5 ); # Get bearings..

    if ( SAPNW04::CleanIPC ( $sSAPAdmin,
                             $sEnvFile,
                             $sSAPSID,
                             $sInstName,
                             $iTimeout ) ) {
      &logTrace ( 20065,
      "<%s> Executed cleanipc successfully", $sMyName );
    }
    else {
      $iReturnCode = 0; # Failure..
      &logError ( 20066,
      "<%s> Failed to execute cleanipc. Returning with [%s]",
      $sMyName, $iReturnCode );
    }
  }
  else {
    &logTrace ( 20067,
    "<%s> Skipping cleanipc for InstType [%s].
    Possibility of destroying active lock table", $sMyName, $sInstType );
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------
  # Initiate SAP log..
  #-------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sMessage = "Shutdown completed [$sResourceName]";

  if ( SAPNW04::UpdateSAPLog ( $sInstName,
                               $sInstType,
                               $sSAPAdmin,
                               $sSAPSID,
                               $sStartProfile,
                               $sMessage ) ) {
    &logTrace ( 20068,
    "<%s> Updated log successfully", $sMyName );
  }
  else {
    &logError ( 20069,
    "<%s> Could not log successfully", $sMyName );
  }

  SAPNW04::DeletePidFile ( $sResourceName );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN String InstName,
                        IN String InstType,
                        IN String SAPAdmin,
                        IN String SAPSID,
                        IN String EnvFile,
                        IN String ProcMon,
                        IN String StartProfile
                        IN Integer CleanCode );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sInstName         String identifier that uniquely identifies a SAP server instance
 $sInstType         String identifier that classifies the SAP server instance type
 $sSAPAdmin         UNIX user dedicated for this instance of SAP
 $sSAPSID           SAP system name defined during installation
 $sEnvFile          Full path to the environment file for SAP
 $sProcMon          List of SAP processes to monitor
 $sStartProfile     Specifies the full path to the start-profile for this instance
 $iCleanCode        Why clean was called
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstName,
       $sInstType,
       $sSAPAdmin,
       $sSAPSID,
       $sEnvFile,
       $sProcMon,
       $sStartProfile,
       $iCleanCode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanCode } ||         # Reason..
                    $iCleanCode ) =~ s/_/ /g;

  &logTrace ( 20070, "Subroutine <%s> called with:
  InstName           [%s]
  InstType           [%s]
  SAPAdmin           [%s]
  SAPSID             [%s]
  EnvFile            [%s]
  ProcMon            [%s]
  StartProfile       [%s]
  Clean reason       [%s]",

  $sMyName,   $sInstName,   $sInstType,    $sSAPAdmin,    $sSAPSID,
  $sEnvFile, $sProcMon, $sStartProfile, $sCleanReason );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my $iSleepTime = 10; # Secs..
  my ( $iTillWhen, $iTimeout, $sKillFile, $iSapStartPid,
  $sCollLockFile, $sSendLockFile, $sMessage, @lsFiles, @liPids ) = ();

  unless ( $hsCleanReasons { $iCleanCode } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanCode } eq 'Offline_was_ineffective' ) {

    &logInfo ( 20071,
    "<%s> Attempting to shutdown SAP Instance [%s] normally",
    $sMyName, $sInstName );

    #---------------
    # Get SAP PIDs..
    #---------------
    SAPNW04::GetProcState ( $sInstName,
                            $sProcMon,
                            $sSAPAdmin,
                            $sSAPSID,
                            $sStartProfile,
                            \@liPids,
                            0 ); # Regular..

    @liPids = $bSapStart ? shift @liRunningProcs : @liRunningProcs; # Global..

    if( $bSapStart && defined $iCCMSX ) {
      push ( @liPids, $iCCMSX );
    }

    if( $bSapStart && defined $iCCMSR ) {
      push ( @liPids, $iCCMSR );
    }

    if ( scalar @liPids ) {
      &logInfo ( 20072, "<%s> Sending SIGINT to PIDs [%s]",
      $sMyName, join ( ' ', @liPids ) );

      #-----------------------------------------
      # Give SAP chance to shutdown gracefully..
      #-----------------------------------------
      if ( kill ( 2, @liPids ) == scalar @liPids ) {
        &logInfo ( 20073,
        "<%s> Sucessfully sent SIGINT to PIDs [%s]. Waiting for [%s] sec",
        $sMyName, join ( ' ', @liPids ), $iSleepTime );

        sleep $iSleepTime; # Sleep..
      }
      else {
        &logError ( 20074,
        "<%s> Failed to give SIGINT to PIDs [%s]. Proceeding with SIGKILL",
        $sMyName, join ( ' ', @liPids ) );
      }
    }
    else {
      &logError ( 20075,
      "<%s> Proceeding with SIGKILL for SAP Instance [%s]",
      $sMyName, $sInstName );
    }
  }

  #----------------------------
  # Remove existing processes..
  #----------------------------
  if ( SAPNW04::KillSAPProcs ( $sInstName,
                               $sProcMon,
                               $sSAPAdmin,
                               $sSAPSID,
                               $sStartProfile ) ) {
    &logInfo ( 20076,
    "<%s> Confirmed no processes exist for SAP instance [%s]",
    $sMyName, $sInstName );
  }
  else {
    &logError ( 20077,
    "<%s> Persistent processes exist for SAP instance [%s]",
    $sMyName, $sInstName );
  }

  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( $sInstType =~ /^(central|dialog|enqrep|aenqrep|jenqrep)$/io ) {

    #--------------------
    # Get time bearings..
    #--------------------
    $iTillWhen = $^T + SAPNW04::GetEngineTimeout ( 1 ); # With margin..
    $iTimeout = $iTillWhen - time;

    #----------------------
    # Clean IPC resources..
    #----------------------
    if ( SAPNW04::CleanIPC ( $sSAPAdmin,
                             $sEnvFile,
                             $sSAPSID,
                             $sInstName,
                             $iTimeout ) ) {
      &logInfo ( 20078,
      "<%s> Executed cleanipc successfully", $sMyName );
    }
    else {
      &logError ( 20079, "<%s> Failed to execute cleanipc.", $sMyName );
      $iReturnCode = 0;
    }
  }
  else {
    &logTrace ( 20080,
    "<%s> Skipping cleanipc for InstType [%s].", $sMyName, $sInstType );
  }

  #------------------
  # Remove kill.sap..
  #------------------
  $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

  if ( -e $sKillFile ) {
    &logTrace ( 20081,
    "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
    push ( @lsFiles, $sKillFile );
  }

  #-----------------------------
  # Remove collector lock file..
  #-----------------------------
  $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

  if ( -e $sCollLockFile ) {
    &logTrace ( 20082,
    "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
    push ( @lsFiles, $sCollLockFile );
  }

  #--------------------------
  # Remove sender lock file..
  #--------------------------
  $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

  if ( -e $sSendLockFile ) {
    &logTrace ( 20083,
    "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
    push ( @lsFiles, $sSendLockFile );
  }

  #-------------------
  # Remove the files..
  #-------------------
  if ( scalar @lsFiles &&
       ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
    &logError ( 20084, "<%s> Failed to remove stale file(s) [%s]",
    $sMyName, join ( ' ', @lsFiles ) );
  }

  #-------------------
  # Initiate SAP log..
  #-------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };

  $sMessage = "Clean called for [$sResourceName] [$sCleanReason]";

  if ( SAPNW04::UpdateSAPLog ( $sInstName,
                               $sInstType,
                               $sSAPAdmin,
                               $sSAPSID,
                               $sStartProfile,
                               $sMessage ) ) {
    &logTrace ( 20085,
    "<%s> Updated log successfully", $sMyName );
  }
  else {
    &logError ( 20086,
    "<%s> Could not log successfully", $sMyName );
  }

  SAPNW04::DeletePidFile ( $sResourceName );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String InstName,
                                   IN String ProcMon,
                                   IN String SAPAdmin,
                                   IN String SAPSID,
                                   IN String StartProfile );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sInstName                 String identifier that uniquely identifies a SAP server instance
 $sProcMon                  List of SAP processes to monitor
 $sSAPAdmin                 UNIX user dedicated for this instance of SAP
 $sSAPSID                   SAP system name defined during installation
 $sStartProfile             Specifies the full path to the start-profile for this instance
 

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstName,
       $sProcMon,
       $sSAPAdmin,
       $sSAPSID,
       $sStartProfile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20087, "Subroutine <%s> called with:
  InstName                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  StartProfile             [%s]",

  $sMyName,
  $sInstName, $sProcMon, $sSAPAdmin, $sSAPSID, $sStartProfile, );

  my ( $sHost, @liDummyPids ) = ();

  ( $sHost = $sStartProfile ) =~ s#.*/.*?_[^_]+_##;

  my $iReturnCode = SAPNW04::GetProcState ( $sInstName,
                                            $sProcMon,
                                            $sSAPAdmin,
                                            $sSAPSID,
                                            $sStartProfile,
                                            \@liDummyPids,
                                            1 ); # Be fast..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20088,
    "<%s> Found all required processes", $sMyName );
  }
  else {
    &logTrace ( 20089,
    "<%s> Did not find all required processes, returning [%s]",
    $sMyName, $iReturnCode );
  }

  &logTrace ( 20090, "<%s> Returning [%s]", $sMyName, $iReturnCode );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( IN String InstName,
                               IN String ProcMon,
                               IN String SAPAdmin,
                               IN String SAPSID,
                               IN String StartProfile,
                               OUT ListReference OnlinePids,
                               IN Boolean UseFFLM );

 Description : Performs the actual first level monitoring

 Parameters :

 $sInstName        String identifier that uniquely identifies a SAP server instance
 $sProcMon         List of SAP processes to monitor
 $sSAPAdmin        UNIX user dedicated for this instance of SAP
 $sSAPSID          SAP system name defined during installation
 $sStartProfile    Specifies the full path to the start-profile for this instance
 $rliOnlinePids    PIDs of relevant running processes
 $bUseFFLM         Flag to indicate if fast FLM is to be attempted

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  PARTIAL                                  VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstName,
       $sProcMon,
       $sSAPAdmin,
       $sSAPSID,
       $sStartProfile,
       $rliOnlinePids,
       $bUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20091, "Subroutine <%s> called with:
  InstName          [%s]
  ProcMon           [%s]
  SAPAdmin          [%s]
  SAPSID            [%s]
  StartProfile      [%s]
  rliOnlinePids     [%s]
  UseFFLM           [%s]",

  $sMyName,
  $sInstName, $sProcMon, $sSAPAdmin, $sSAPSID,
  $sStartProfile,  $rliOnlinePids,  $bUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iUseRFLM,      $iFound,      $iPid,      $iParent,     $iJcFlag,
       $iProcessCount,       $iPosn,      $iSleepTime,       $iDwMaster,
       $iSapstartPid,      $bAddIn,     $sKey,   $sTemp,      $sPidFile,
       $sProc,          $sOSVer,       $sInstProfile,            $sLang,
       $sEntryPointName,   $sSwitch, $rhlProc,  @lsProcMon,   @lsValues,
       @lsUpdates,      @hProcListHash,     %hiDwPids,      %hiProcPids,
       %hsProcMonDetails, %hsProcMonCopy ) = ();

  my $iUid = getpwnam $sSAPAdmin; # Should not fail..

  #---------------------------------------------
  # Derive instance profile from start profile..
  #---------------------------------------------
  ( $sInstProfile = $sStartProfile ) =~ s/START_/${sSAPSID}_/;
  my @lsOthers = ( "/usr/sap/$sSAPSID/$sInstName/igs",
                   "pf=.*$sSAPSID.*$sInstName.*", "${sSAPSID}_${sInstName}",
                   'sapstartsrv\s+' . "pf=.*$sSAPSID.*$sInstName.*");

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  #-------------------------
  # Check if we are ADD-IN..
  #-------------------------
  ++ $bAddIn if ( 2 == grep { /^(jc|dw)$/o } $sProcMon =~ /([^\s]+)/g );

  #-----------------------------------------------
  # Insert sapstart, unless its a recursive call..
  #-----------------------------------------------
  if ( $sProcMon eq 'jlaunch' ) {
    $sProc = $sProcMon;
  }
  elsif ( $sProcMon eq 'JC' ) {
    $sProc = 'jc dw'; # ADD-IN for recursive call..
  }
  else {
    $sProc =  "sapstart $sProcMon";
  }

  @$rliOnlinePids = (); # Truncate..

  #-------------------------------------
  # Get details for each procmon value..
  #-------------------------------------
  if ( SAPNW04::GetProcMonDetails ( $sInstName,
                                    $sSAPSID,
                                    $sProc,
                                    $sStartProfile,
                                    \%hsProcMonDetails ) ) {

    delete $hsProcMonDetails { dw } if $sProcMon eq 'JC'; # Remove dw in recursive..

    $iUseRFLM = scalar keys %hsProcMonDetails; # Set to entire count..
    %hsProcMonCopy = %hsProcMonDetails;        # Backup..

    #---------------------------------
    # Check if we attempt to be fast..
    #---------------------------------
    if ( $sEntryPointName ne 'offline' && $bUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( sort keys %hsProcMonCopy ) { # Use copy..
        @lsUpdates = ();                           # Reset..
        @lsValues = @{ $hsProcMonCopy { $sKey } }; # Get values..
        $sTemp .= "\nSAP process  [$sKey]\n";      # Display string..

        while ( ( $sPidFile, $sProc ) =
                  splice ( @lsValues, 0, 2 ) ) {

          #---------------
          # Run fast FLM..
          #---------------
          $iFound = SAPNW04::FFLMStatus ( $sPidFile,
                                          $sProc,
                                          $iUid,
                                          \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsProcMonCopy { $sKey } } = @lsUpdates;

        #---------------------
        # Handle exit status..
        #---------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          if ( $sKey eq 'sapstart' ) {
            $iSapstartPid = $iPid; # Store..
            unshift ( @$rliOnlinePids, $iPid ); # Insert at the front..
          }
          else {
            if ( $sKey eq 'dw' ) {
              $iDwMaster = $iPid; # Store master..
              &logTrace ( 20092,
              "<%s> Found master dw process PID [%s] based on PID file",
              $sMyName, $iDwMaster );
            }
            elsif ( $sKey eq 'jc' ) {
              ++ $iJcFlag; # Set flag..
            }

            push ( @$rliOnlinePids, $iPid ); # Insert at the end..
          }
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      &logTrace ( 20093,
      "<%s> SapInst [%s] ProcMon details ---%s\n\n",
      $sMyName, $sInstName, $sTemp );
 
      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use regular FLM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( sort keys %hsProcMonCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsProcMonCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsProcMonCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else {
    &logError ( 20094,
    "<%s> Failed to get details with ProcMon [%s]",
    $sMyName, $sProcMon );
    return VCSDef_MonitorExit::UNKNOWN;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) { 
    &logTrace ( 20095,
    "<%s> Running regular first level monitor on [%s] processes",
    $sMyName, $iUseRFLM );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20096,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    } 
    else {
      &logError ( 20097,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-------------------
    # Solaris 10 zones..
    #-------------------
    $iParent = 1; # Set to init..
    $sOSVer = $^O eq 'solaris' ? ( uname ) [2] : '';

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      &logTrace ( 20098, "<%s> Searching PID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { COMMAND } );

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      $iPid = $$rhlProc { PID }; # Assign..

      #--------------------------------
      # Handle Solaris 10 local zones..
      #--------------------------------
      if ( $sOSVer eq '5.10'                       && # Solaris 10..
           $$rhlProc { COMMAND } =~ m@/sbin/init$@ && # With init..
           ! kill ( 0, 1 ) ) {                        # In a non-global zone..
        $iParent = $$rhlProc { PPID };                # Get actual PID..
      }

      #-----------------------------------------------
      # Look for processes that were offline in FFLM..
      #-----------------------------------------------
      foreach $sKey ( sort keys %hsProcMonCopy ) {
        @lsUpdates = (); # Reset..
        @lsValues = @{ $hsProcMonCopy { $sKey } }; # Get values..

        #--------------------------------------------------
        # Should have details per three elements of array..
        # <PIDFILE> <PROCESS> <STATE>
        #--------------------------------------------------
        while ( ( $sPidFile, $sProc, $iFound ) =
                  splice ( @lsValues, 0, 3 ) ) {

          #--------------------------------------
          # Verify if process is really offline..
          #--------------------------------------
          if ( VCSDef_MonitorExit::OFFLINE == $iFound ) {  # Earlier marked offline..
            if ( $$rhlProc { UID } == $iUid &&             # Owned by SAPAdmin..
                 $$rhlProc { COMMAND } =~ /\b$sProc\b/ ) { # Was now found..

              #-----------------------
              # Handle dw separately..
              #-----------------------
              $iFound = VCSDef_MonitorExit::ONLINE unless $sKey eq 'dw';

              &logTrace ( 20099,
              "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
              $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND }, $sProc ); 

              #-----------------------
              # Handle special cases..
              #-----------------------
              $sSwitch = $sKey =~ /^(jc|jlaunch|dw|sapstart)$/o ? $sKey:
                                                                  'default';
              goto $sSwitch; {
                sapstart:
                  $iPosn = 0;
                  $iSapstartPid = $iPid; # Store..
                last;

                jc:
                  $iPosn = 1;
                  ++ $iJcFlag; # Set flag..
                last;

                jlaunch:
                  $iPosn = 2;
                last;

                dw:
                  $iPosn = 3;
                  $hiDwPids { $iPid } = $$rhlProc { PPID }; # Store..
                last;

                default:
                  $iPosn = scalar @$rliOnlinePids;
                last;
              }

              #--------------------------------
              # Insert into appropriate place..
              #--------------------------------
              $iPosn = scalar @$rliOnlinePids if $iPosn > scalar @$rliOnlinePids;

              splice ( @$rliOnlinePids, $iPosn, 0, $iPid ); # Insert..
              $hiProcPids { $sKey } = $iPid;                # Store PID..

              &logTrace ( 20100,
              "<%s> Stored PID [%s] for proc [%s] at index [%s]",
              $sMyName, $iPid, $sKey, $iPosn );
            }
          }

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );
        }

        @{ $hsProcMonCopy { $sKey } } = @lsUpdates; # Update..
      }

      #---------------------------------
      # Handle other related processes..
      #---------------------------------
      if ( ( ! grep { /^$iPid$/ } @$rliOnlinePids ) &&         # Not caught earlier..
           $$rhlProc { UID } == $iUid               &&         # Owned by SAPAdmin..
           $sProcMon !~ /^(JC|jlaunch)$/o           &&         # Not in recursive call..
           map { $$rhlProc { COMMAND } =~ /$_/ } @lsOthers ) { # Was now found..
        push ( @$rliOnlinePids, $iPid );                       # Store..

        #-----------------------------------------
        # Populate dwindling sapstartsrv process..
        #-----------------------------------------
        if ( $$rhlProc { COMMAND } =~ /sapstartsrv/o ) {
          $iSapStartSrv = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /cm\.sap/o ) { # Note down sapccm4x process
          $iCCMSX = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /cs\.sap/o ) { # Note down sapccmsr process
          $iCCMSR = $iPid; # Populate global..
        }

        &logTrace ( 20101,
        "<%s> Matched process PID [%s] PPID [%s] CMD [%s]",
        $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $i, $bSapstartFlag, $iOnline, $iOffline, @liDummyPids ) = ();

  my @liDwPids = keys %hiDwPids;
  $sTemp = ''; # Ensure..

  foreach $sKey ( sort keys %hsProcMonCopy ) {
    @lsValues = @{ $hsProcMonCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {

      #----------------------
      # Set status for 'dw'..
      #----------------------
      if ( $sKey eq 'dw' && scalar @liDwPids ) {
        #----------------------------
        # Check if we the master dw..
        #----------------------------
        if ( $iSapstartPid &&                    # Sapstart exists..
             ( ( $iDwMaster ) =                  # Get the master dw..
             grep {                              # By finding..
               $iSapstartPid == $hiDwPids { $_ } # sapstart as its parent..
             } @liDwPids ) ) {                   # From all dw processes..
          $hiProcPids { $sKey } = $iDwMaster;    # Store..
          $iFound = VCSDef_MonitorExit::ONLINE;  # And mark online..
          &logTrace ( 20102,
          "<%s> Found master dw process PID [%s] PPID [%s]", 
          $sMyName, $iDwMaster, $iSapstartPid );
        }
        elsif ( ( ( $iDwMaster ) =               # A dw process..
                grep {                           # Found..
                  $iParent == $hiDwPids { $_ } &&# With init as parent..
                  ++ $i                          # Mark it..
                } @liDwPids ) && 1 == $i ) {     # If only one such exists..
          $hiProcPids { $sKey } = $iDwMaster;    # Store..
          $iFound = VCSDef_MonitorExit::ONLINE;  # And mark online..
          &logWarn ( 20103,
          "<%s> Missing sapstart process.
          Found master dw process PID [%s] PPID [%s]",
          $sMyName, $iDwMaster, $hiDwPids { $iDwMaster } );
        }                                              
        else {                                   # Error..
          &logTrace ( 20104,
          "<%s> Missing dw master. Marking status as offline", $sMyName );

          $iFound = VCSDef_MonitorExit::OFFLINE; # Ensure its offline..

          #------------------------
          # Delete PID, if stored..
          #------------------------
          if ( defined $hiProcPids { $sKey } ) {
            delete $hiProcPids { $sKey };
          }
        }
      }

      #--------------------------
      # Mark status of sapstart..
      #--------------------------
      if ( $sKey eq 'sapstart' &&
           $iSapstartPid ) {
        ++ $bSapstartFlag; # Mark it found..
        ++ $bSapStart;     # Mark global..
      }

      #--------------------------------
      # If we are ADD-IN, validate jc..
      #--------------------------------
      if ( $sKey eq 'jc'                 && # Looking for jc..
           $sEntryPointName eq 'monitor' && # in monitor..
           $bAddIn                       && # dw and jc declared..
           $iDwMaster ) {                   # dw master exists..
        if ( ! $iJcFlag ) {                 # jc missing..
          #-----------------------------------
          # Calculate sleep time dynamically..
          #-----------------------------------
          $iSleepTime = SAPNW04::GetEngineTimeout ( 0 ) / 10;

          &logWarn ( 20105,
          "<%s> Missing process [%s], though 'dw' master [%s] exists.
          Retrying in [%s] sec", $sMyName, $sKey, $iDwMaster, $iSleepTime );

          sleep $iSleepTime; # Take a break..

          #---------------------------
          # Call ourselves with 'JC'..
          #---------------------------
          $iFound = SAPNW04::GetProcState ( $sInstName,
                                            'JC',           # In CAPS..
                                            $sSAPAdmin,
                                            $sSAPSID,
                                            $sStartProfile,
                                            \@liDummyPids,
                                            0 );            # Regular FLM..

          #--------------------------------------------------------------------
          # If we returned online, we know we found jc and at least one jaunch.
          # PID of jc now is in first position..
          #--------------------------------------------------------------------
          if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
              &logTrace ( 20106,
              "<%s> Found jc process with PID [%s] after recursive call",
              $sMyName, $liDummyPids [0] );

              $hiProcPids { $sKey } = shift @liDummyPids; # Store..
              ++ $iJcFlag; # Set flag..

              #-----------------
              # Update jlaunch..
              #-----------------
              $hiProcPids { jlaunch } = shift @liDummyPids;   # Store..
              ${ $hsProcMonCopy { jlaunch } } [-1] = $iFound; # Mark online..
          }
          else {
            &logError ( 20107, "<%s> Missing jc after recursive call", $sMyName );
          }
        }
      }

      #--------------------------------------------------
      # Check if 'jlaunch' is missing and 'jc' is alive..
      #--------------------------------------------------
      if ( $sEntryPointName eq 'monitor' &&           # In monitor..
           $sKey eq 'jlaunch'            &&           # Jaunch,
           $iJcFlag                      &&           # when jc, alive..
           VCSDef_MonitorExit::OFFLINE == $iFound ) { # is not..
        #-----------------------------------
        # Calculate sleep time dynamically..
        #-----------------------------------
        $iSleepTime = SAPNW04::GetEngineTimeout ( 0 ) / 10;

        &logWarn ( 20108,
        "<%s> Missing process [%s], though 'jc' exists. Retrying in [%s] sec",
        $sMyName, $sKey, $iSleepTime );

        sleep $iSleepTime; # Take a break..

        #--------------------------------------
        # Call ourselves again with 'jlaunch'..
        #--------------------------------------
        $iFound = SAPNW04::GetProcState ( $sInstName,
                                          'jlaunch',
                                          $sSAPAdmin,
                                          $sSAPSID,
                                          $sStartProfile,
                                          \@liDummyPids,
                                          0 ); # Regular FLM..
        #--------
        # Store..
        #--------
        if ( scalar @liDummyPids ) {
          push ( @$rliOnlinePids, @liDummyPids );
          $hiProcPids { $sKey } = $liDummyPids [0];

          &logTrace ( 20109,
          "<%s> Found jlaunch process with PID [%s] after another attempt",
          $sMyName, $liDummyPids [0] );
        }
      } 

      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                  && # In RFLM..
           $sEntryPointName =~ /^(monitor|online)$/o  && # In monitor or online..
           $hiProcPids { $sKey }                      && # PID populated..
           SAPNW04::CreatePidFile ( $sPidFile,           # Create file..
                                    $hiProcPids { $sKey } ) ) {
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound &&
              $sKey ne 'sapstart') { # Ignore sapstart..
        $iOffline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOnline && $iOffline?VCSDef_MonitorExit::PARTIAL_ONLINE:
                 $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;
    &logTrace ( 20110,
    "<%s> Found [%s] PIDs [%s] for SAP instance. Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ), $iReturnCode );
  }

  &logTrace ( 20111,
  "<%s> SAP Instance [%s] ProcMon details ---\n%s
  Return code is [%s]", $sMyName, $sInstName, $sTemp, $iReturnCode );

  $sTemp = undef; # Release..

  #---------------------
  # Report on sapstart..
  #---------------------
  if ( ! $bSapstartFlag               &&               # Missing sapstart..
       $sEntryPointName eq 'monitor'  &&               # In monitor..
       $sProcMon !~ /^(JC|jlaunch)$/o &&               # Not in recursive call..
       VCSDef_MonitorExit::OFFLINE != $iReturnCode ) { # Not offline..

    &logWarn ( 20112,
    "<%s> Missing sapstart process in first level monitor",
    $sMyName );
  } 
       
  #-------------------------------- 
  # Populate the global PID array..
  #-------------------------------- 
  @liRunningProcs = @$rliOnlinePids;

  &logTrace ( 20113,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20114, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20115, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20116, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20117, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20118,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );

  }
  else {
    &logError ( 20119,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20120,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( SAPNW04::GetPidDetails ( $iPid,
                                  \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20121,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&           # Owner match..
               $$rhlProc { COMMAND } =~ /\b$sProc\b/ ) { # Filter match..
          &logTrace ( 20122,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20123,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20124,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20125, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetProcMonDetails ( IN String InstName
                                     IN String SAPSID
                                     IN String ProcMon
                                     IN String StartProfile
                                     OUT RefHashList ProcMonDetails );

 Description: Gives PID file and process string for each ProcMon value

 Parameters :

 $sInstName           String identifier that uniquely identifies a SAP server instance
 $sSAPSID             SAP system name defined during installation
 $sProcMon            List of SAP processes to monitor
 $sStartProfile       Specifies the full path to the start-profile
 $rhlProcMonDetails   Stores details of PID file and process string

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetProcMonDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstName,
       $sSAPSID,
       $sProcMon,
       $sStartProfile,
       $rhlProcMonDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20126, "Subroutine <%s> called with:
  InstName                 [%s]
  SAPSID                   [%s]
  ProcMon                  [%s]
  StartProfile             [%s]
  rhlProcMonDetails        [%s]",

  $sMyName,
  $sInstName, $sSAPSID, $sProcMon, $sStartProfile, $rhlProcMonDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sProc, $sTmpProc, $sPidFile, $sInstProfile, %hsProcMon, $sHost ) = ();

  my $sArch = 0;
  my %hsArch = ( 0 => '',
                 1 => 'JAVA',
                 2 => 'ABAP',
                 3 => 'ADDIN' );

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  map {
    $sArch |= 1 if /^jc$/;
    $sArch |= 2 if /^dw$/;
  } @lsProcMon;

  #----------------------------
  # Insert jlaunch, if needed..
  #----------------------------
  push @lsProcMon, 'jlaunch' if ( $sArch & 1 ); # jc located..

  my $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #---------------------------------------------
  # Derive instance profile from start profile..
  #---------------------------------------------
  ( $sInstProfile = $sStartProfile ) =~ s/START_/${sSAPSID}_/;

  #---------------------------------------------
  # Derive virtual hostname from start profile..
  #---------------------------------------------
  ( $sHost = $sStartProfile ) =~ s#.*/.*?_[^_]+_##; # Get hostname..

  %$rhlProcMonDetails = (); # Truncate..

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20127, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20128, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  #-----------------------------------------------
  # Create PID file and process strings for each..
  #-----------------------------------------------
  foreach $sProc ( @lsProcMon ) {
    $sPidFile = "$sPIDDir/pidfile.$sProc";

    #----------------------------------------
    # Handle sapstart and ADDIN differently..
    #----------------------------------------
    if ( $sProc eq 'sapstart' ) {
      $sTmpProc = "$sProc pf=.*$sSAPSID.*$sInstName.*$sHost";
      #$sTmpProc = "$sProc pf=$sStartProfile";
    }
    elsif ( $sProc eq 'jc'     &&              # jc defined..
            $hsArch { $sArch } eq 'ADDIN' ) {  # in ADDIN..
      $sTmpProc = "jcontrol pf=.*$sSAPSID.*$sInstName.*$sHost"; # look for jcontrol..
    }
    elsif ( $sProc eq 'jlaunch' ) {
      $sTmpProc = "$sProc pf=.*$sSAPSID.*$sInstName.*$sHost.*server\\d";
    }
    elsif( $sProc eq 'gwrd' || $sProc eq 'icman' ){
      $sTmpProc = "$sProc.*pf=.*$sSAPSID.*$sInstName.*$sHost";
    }
    else {
      $sTmpProc = "$sProc.sap${sSAPSID}_${sInstName}.*pf=.*$sSAPSID.*$sInstName.*$sHost";
      #$sTmpProc = "$sProc.sap${sSAPSID}_${sInstName}.*pf=$sInstProfile";
    }

    &logTrace ( 20129,
    "<%s> Process [%s] Search Pattern [%s] PID file [%s]",
     $sMyName, $sProc, $sTmpProc, $sPidFile );

    push @{ $hsProcMon { $sProc } }, $sPidFile, $sTmpProc;
  }

  %$rhlProcMonDetails = %hsProcMon; # Now store..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String InstName,
                          IN String InstType,
                          IN String MonitorProgram,
                          IN String SAPAdmin,
                          IN String ProcMon
                          IN String SAPMonHome,
                          IN String SAPSID,
                          IN String StartProfile,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile            Full path to the environment file for SAP
 $sInstName           String identifier that uniquely identifies a SAP server instance
 $sInstType           String identifier that classifies the SAP server instance type
 $sMonitorProgram     Full path to external monitor program
 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sProcMon            List of SAP processes to monitor
 $sSAPMonHome         Directory defining the location of the SAP monitoring utilities
 $sSAPSID             SAP system name defined during installation
 $sStartProfile       Specifies the full path to the start-profile
 $iSecondLevelMonitor Cycle count to run in-depth monitoring
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstName,
       $sInstType,
       $sMonitorProgram,
       $sSAPAdmin,
       $sProcMon,
       $sSAPMonHome,
       $sSAPSID,
       $sStartProfile,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20130, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstName                 [%s]
  InstType                 [%s]
  MonitorProgram           [%s]
  SAPAdmin                 [%s]
  ProcMon                  [%s]
  SAPMonHome               [%s]
  SAPSID                   [%s]
  StartProfile             [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,
  $sEnvFile,  $sInstName,  $sInstType,  $sMonitorProgram,
  $sSAPAdmin, $sProcMon, $sSAPMonHome, $sSAPSID, $sStartProfile,
  $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------------------------------------
  # Upper bound..
  # Time in EPOCH when entry pt started + VCS time limit..
  #-------------------------------------------------------
  my $iTillWhen = $^T + SAPNW04::GetEngineTimeout ( 1 ); # With margin..

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
      SAPNW04::DelayedAttributeCheck ( $sEnvFile,
                                       $sSAPMonHome,
                                       $sStartProfile ) ) ) {
      &logTrace ( 20131,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20132,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20133,
    "<%s> Proceeding with second level monitoring", $sMyName );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPNW04::SecondLevelMonitor ( $sEnvFile,
                                       $sInstName,
                                       $sInstType,
                                       $sSAPAdmin,
                                       $sProcMon,
                                       $sSAPMonHome,
                                       $sSAPSID,
                                       $sStartProfile,
                                       $iTillWhen ) ) ) {
      &logTrace ( 20134,
      "<%s> SAP Instance [%s] online with status [%s] after SLM",
      $sMyName, $sInstName, $iReturnCode );
    }
    else {
      &logError ( 20135,
      "<%s> SAP Instance [%s] NOT online. Status [%s]",
      $sMyName, $sInstName, $iReturnCode );
      return $iReturnCode; # Get out..
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         SAPNW04::ExternalMonitor ( $sSAPAdmin,    # after execution..
                                    $sEnvFile,
                                    $sExtCmd,
                                    $sExtCmdArgs,
                                    $iTillWhen ) ) ) {
      &logTrace ( 20136,
      "<%s> SAP Instance [%s] found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $sInstName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20137,
      "<%s> SAP Instance [%s] NOT online with external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sInstName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ChkResName ( IN String ResName,
                              IN String SAPSID,
                              IN String InstType );

 Description : Checks if resource has been configured

 Parameters :

 $sResName         Resource Name to check
 $sSAPSID          SAP system name defined during installation
 $sInstType        String identifier that classifies the SAP server instance type

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ChkResName {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sResName,
       $sSAPSID,
       $sInstType ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20138, "Subroutine <%s> called with:
  ResName            [%s]
  SAPSID             [%s]
  InstType           [%s]",

  $sMyName, $sResName, $sSAPSID, $sInstType );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------------
  # Validate string for resource..
  #-------------------------------
  return 0 unless ( $sResName =~ /\S/o ); # Must be a non-white string..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..
  my $sCmdArgs = "-display $sResName -attribute InstType SAPSID";

  #-----------------------
  # Define valid matches..
  #-----------------------
  my %hsInstMatch = ( 'ENQREP'  => 'ENQUEUE',
                      'AENQREP' => 'AENQUEUE',
                      'JENQREP' => 'JENQUEUE' );

  my ( $iExitCode, @lsCmd ) = ();

  &logTrace ( 20139,
  "<%s> Executing Cmd [%s] with Args [%s]",
   $sMyName, $sCmd, $sCmdArgs );

  #--------------------------
  # Fire command and verify..
  #--------------------------
  ( @lsCmd ) = grep {
                 ( s/\n// || 1 )   &&            # Chomp, and make true..
                 /InstType|SAPSID/ &&            # Contains these..
                 s/.*\s+//                       # Get last field..
               } readpipe ( "$sCmd $sCmdArgs" ); # From output..

  #--------------------
  # Collect exit code..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20140,
    "<%s> Cmd [%s] with Args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ( 2 == scalar @lsCmd )                  &&          # Got both values..
          ( $sSAPSID eq $lsCmd [-1] )             &&          # SAPSID is a match..
          defined $hsInstMatch { uc $sInstType }  &&          # InsType is defined..
          $hsInstMatch { uc $sInstType } eq uc $lsCmd [0] ) { # And a correct pair..
    &logTrace ( 20141,
    "<%s> Resource [%s] defined in cluster with InstType [%s] SAPSID [%s]",
     $sMyName, $sResName, $lsCmd [0], $lsCmd [-1] );
  }
  else {
    &logError ( 20142,
    "<%s> Resource [%s] not a matching ENQUEUE variant for this resource",
    $sMyName, $sResName );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : Int DelayedAttributeCheck ( IN String EnvFile,
                                        IN String SAPMonHome,
                                        IN String StartProfile );

 Description : Just In Time Validation

 Parameters :

 $sEnvFile            Full path to the environment file for SAP
 $sSAPMonHome         Directory defining the location of the SAP monitoring utilities
 $sStartProfile       Specifies the full path to the start-profile


 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sSAPMonHome,
       $sStartProfile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20143, "Subroutine <%s> called with:
  EnvFile         [%s]
  SAPMonHome      [%s]
  StartProfile    [%s]",

  $sMyName, $sEnvFile, $sSAPMonHome, $sStartProfile );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-----------------------------------------
  # Validate each of the attribute in turn..
  #-----------------------------------------

  #-----------------------
  # EnvFile has to exist..
  #-----------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20144,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
  }

  #-------------------
  # Validate SAPHome..
  #-------------------
  if ( ! -d $sSAPMonHome ) {
    &logError ( 20145,
    "<%s> SAPMonHome [%s] not a valid directory", $sMyName, $sSAPMonHome );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
  }

  #----------------------------
  # Startprofile has to exist..
  #----------------------------
  if ( ! -r $sStartProfile ) {
    &logError ( 20146,
    "<%s> StartProfile does not exist [%s]", $sMyName, $sStartProfile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool UpdateSAPLog ( IN String InstName,
                                IN String InstType,
                                IN String SAPAdmin,
                                IN String SAPSID,
                                IN String StartProfile,
                                IN String MessageText );

 Description : Update SAP log with passed message string..

 Parameters :

 $sInstName         String identifier that uniquely identifies a SAP server instance
 $sInstType         String identifier that classifies the SAP server instance type
 $sSAPAdmin         UNIX user dedicated for this instance of SAP
 $sSAPSID           SAP system name defined during installation
 $sStartProfile     Specifies the full path to the start-profile for this instance
 $sMessageText      Message text for update


 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub UpdateSAPLog {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstName,
       $sInstType,
       $sSAPAdmin,
       $sSAPSID,
       $sStartProfile,
       $sMessageText ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20147, "Subroutine <%s> called with:
  InstName                 [%s]
  InstType                 [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  StartProfile             [%s]
  MessageText              [%s]",

  $sMyName, $sInstName, $sInstType, $sSAPAdmin, $sSAPSID,
  $sStartProfile, $sMessageText );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sDate = strftime '%H:%M:%S %d %b %Y', localtime;
  my $sHost = ( uname ) [1];
  my $sLogFile = "/usr/sap/$sSAPSID/$sInstName/log/" .
                 "${sSAPSID}_${sInstName}.log";

  my $sCpyHeader = '+' . '-' x 75 . '+';
  my $sCmnt = "|";

  #------------------------
  # Define logging format..
  #------------------------

  format SAPLOG =
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$sCpyHeader
@ Date: @<<<<<<<<<<<<<<<<<<<<< Node: @<<<<<<<<<<<<<<<<<<                    @
$sCmnt, $sDate, $sHost, $sCmnt
@ InstName: @<<<<<<<<<<<< InstType: @<<<<<<<<<< SAPSID: @<<<                @
$sCmnt, $sInstName, $sInstType, $sSAPSID, $sCmnt
@ Start profile: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @
$sCmnt, $sStartProfile, $sCmnt
@ @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  @
$sCmnt, $sMessageText, $sCmnt
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$sCpyHeader
.

  #-----------------------
  # Open file for update..
  #-----------------------
  if ( sysopen ( SAPLOG, $sLogFile, O_CREAT|O_WRONLY|O_APPEND|O_SYNC, 0644 ) ) {
    &logTrace ( 20148,
    "<%s> Opened log file [%s] succesfully for update", $sMyName, $sLogFile );

    write SAPLOG; # Update..
 
    unless ( close ( SAPLOG ) ) {
      &logError ( 20149,
      "<%s> Failed to close log file [%s] after update: [%s]",
      $sMyName, $sLogFile, $! );
      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logError ( 20150,
    "<%s> Failed to open log file [%s] for update: [%s]",
    $sMyName, $sLogFile, $! );
    $iReturnCode = 0; # Failure..
  }

  #----------------
  # Set ownership.. 
  #----------------
  my ( $iUid, $iGid ) = ( getpwnam $sSAPAdmin ) [2, 3]; # Should not fail..

  if ( chown ( $iUid, $iGid, $sLogFile ) ) {
    &logTrace ( 20151,
    "<%s> Set ownership of [%s] to UID [%s] GID [%s]",
    $sMyName, $sLogFile, $iUid, $iGid );
  }
  else {
    &logError ( 20152,
    "<%s> Failed to set ownership of [%s] to UID [%s] GID [%s]: [%s]",
    $sMyName, $sLogFile, $iUid, $iGid, $! );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CleanIPC ( IN String SAPAdmin,
                            IN String EnvFile,
                            IN String SAPSID,
                            IN String InstName,
                            IN Integer Timeout );

 Description : Cleans IPC resources for given SAPSID

 Parameters :
 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sEnvFile            Full path to the environment file for SAP
 $sSAPSID             SAP system name defined during installation
 $sInstName           String identifier that uniquely identifies a SAP server instance
 $iTimeout            Timeout in seconds

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CleanIPC {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sSAPSID,
       $sInstName,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20153, "Subroutine <%s> called with:
  SAPAdmin            [%s]
  EnvFile             [%s]
  SAPSID              [%s]
  InstName            [%s]
  Timeout             [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sSAPSID, $sInstName, $iTimeout );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iInstId, $iExitCode, @lsCmdOut ) = ();
  ( $iInstId ) = $sInstName =~ /(\d{2})$/o; # Get the last two..

  my $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/cleanipc";
  my $sCmdArgs = "$iInstId remove";
  my $sLogFile = "/usr/sap/$sSAPSID/$sInstName/log/" .
                 "${sSAPSID}_${sInstName}.log";


  #---------------
  # Validate Cmd..
  #---------------
  if ( ! -e $sCmd ) {
    &logError ( 20154,
    "<%s> Command [%s] does not exist", $sMyName, $sCmd );

    $sCmd = "/usr/sap/$sSAPSID/$sInstName/exe/cleanipc";
    if ( ! -e $sCmd ) {
      &logError ( 20155,
      "<%s> Command [%s] does not exist", $sMyName, $sCmd );
      $iReturnCode = 0; # Failure..
    }
  }

  if ( ! -x $sCmd ) {
    &logError ( 20156,
    "<%s> Command [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #------------------
  # Execute command..
  #------------------
  &logTrace ( 20157,
  "<%s> Executing Cmd [%s] with args [%s] within [%s] sec",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  if ( SAPNW04::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCmd,
                               $sCmdArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdOut ) ) {
    &logTrace ( 20158,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  elsif ( ! $iExitCode ) {
    &logError ( 20159,
    "<%s> Command [%s] with args [%s] timed out in [%s] sec",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout );
    $iReturnCode = 0; # Failure..
  }
  else {
    &logError ( 20160,
    "<%s> Command [%s] with args [%s] failed with exit code [%s]. Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
    $iReturnCode = 0; # Failure..
  }

  #------------------
  # Update log file..
  #------------------
  if ( scalar @lsCmdOut ) {
    if ( sysopen ( SAPLOG, $sLogFile,
                   O_CREAT|O_WRONLY|O_APPEND|O_SYNC, 0644 ) ) {
      &logTrace ( 20161,
      "<%s> Opened log file [%s] succesfully for update",
      $sMyName, $sLogFile );
  
      #------------------------------
      # Put information for message..
      #------------------------------
      unshift ( @lsCmdOut, "Cleanipc command output:\n" );

      print SAPLOG "$_\n" foreach ( @lsCmdOut ); # Update..
  
      unless ( close ( SAPLOG ) ) {
        &logError ( 20162,
        "<%s> Failed to close log file [%s] after update: [%s]",
        $sMyName, $sLogFile, $! );
      }
    }
    else {
      &logError ( 20163,
      "<%s> Failed to open log file [%s] for update: [%s]",
      $sMyName, $sLogFile, $! );
    }
  } 

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20164, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   10 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  $sCommand = $sCmdHares; # Get global..
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  $sCommandArgs = "-value $sResourceName $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                    =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20165,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
  }
  elsif ( ! $iEngineTimeout ) { # Undefined value..
    &logTrace ( 20166, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
  }
  else { # Success..
    &logTrace ( 20167, "<%s> Got [%s] as [%s] sec from engine",
    $sMyName, $sAttribute, $iEngineTimeout );
  }

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  unless ( $iEngineTimeout ) {
    $sCommand = $sCmdHatype; # Get global..
    ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
    $sResourceType = 'SAPNW04' if $sResourceType !~ /\w/; # Ensure..
    $sCommandArgs = "-value $sResourceType $sAttribute";

    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;

    $iExitCode = $? >> 8;  # Get exit status..

    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode ) {
      &logTrace ( 20168,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
      $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
    }
    elsif ( ! $iEngineTimeout ) { # Empty value..
      &logTrace ( 20169, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
      $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
    }
    else { # Success..
      &logTrace ( 20170, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  }
  else {
      &logTrace ( 20171, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                 # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20172,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillSAPProcs ( IN String InstName,
                                IN String ProcMon,
                                IN String SAPAdmin,
                                IN String SAPSID,
                                IN String StartProfile );

 Description : Kills processes for a given SAP instance..

 Parameters :

 $sInstName       String identifier that uniquely identifies a SAP server instance
 $sProcMon        List of SAP processes to monitor
 $sSAPAdmin       UNIX user dedicated for this instance of SAP
 $sSAPSID         SAP system name defined during installation
 $sStartProfile   Specifies the full path to the start-profile for this instance

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub KillSAPProcs {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstName,
       $sProcMon,
       $sSAPAdmin,
       $sSAPSID,
       $sStartProfile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20173, "Subroutine <%s> called with:
  InstName         [%s]
  ProcMon          [%s]
  SAPAdmin         [%s]
  SAPSID           [%s]
  StartProfile     [%s]",

  $sMyName,
  $sInstName, $sProcMon, $sSAPAdmin, $sSAPSID, $sStartProfile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #----------------------
  # Set local variables..
  #----------------------
  my ( @liGoners, @liUnDead ) = ();

  #-------------------------------------
  # If called during online or offline..
  #-------------------------------------
  if ( ( caller 1 )[3] eq 'SAPNW04::Start' || # Start logic..
       ( caller 1 )[3] eq 'SAPNW04::Stop' ) { # Stop logic..
    @liGoners = @liRunningProcs;              # Get global..
  }
  else {
    #----------------------------
    # Get processes for killing..
    #----------------------------
    SAPNW04::GetProcState ( $sInstName,
                            $sProcMon,
                            $sSAPAdmin,
                            $sSAPSID,
                            $sStartProfile,
                            \@liGoners,
                            0 ); # Use regular..
  }

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20174,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20175,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20176,
    "<%s> Could not locate any process for SAP Instance [%s]",
    $sMyName, $sInstName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EnvFile,
                                     IN String InstName,
                                     IN String InstType,
                                     IN String SAPAdmin,
                                     IN String ProcMon,
                                     IN String SAPMonHome,
                                     IN String SAPSID,
                                     IN String StartProfile,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEnvFile        Full path to the environment file for SAP
 $sInstName       String identifier that uniquely identifies a SAP server instance
 $sInstType       String identifier that classifies the SAP server instance type
 $sSAPAdmin       UNIX user dedicated for this instance of SAP
 $sProcMon        List of SAP processes to monitor
 $sSAPMonHome     Directory defining the location of the SAP monitoring utilities
 $sSAPSID         SAP system name defined during installation
 $sStartProfile   Specifies the full path to the start-profile for this instance
 $iTillWhen       Represents the upper limit set by engine in EPOCH

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstName,
       $sInstType,
       $sSAPAdmin,
       $sProcMon,
       $sSAPMonHome,
       $sSAPSID,
       $sStartProfile,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20177, "Subroutine <%s> called with:
  EnvFile          [%s]
  InstName         [%s]
  InstType         [%s]
  SAPAdmin         [%s]
  ProcMon          [%s]
  SAPMonHome       [%s]
  SAPSID           [%s]
  StartProfile     [%s]
  TillWhen         [%s]",

  $sMyName,
  $sEnvFile, $sInstName, $sInstType, $sSAPAdmin, $sProcMon,
  $sSAPMonHome, $sSAPSID, $sStartProfile, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my %hsArch = ( 1 => 'JAVA',
                 2 => 'ABAP',
                 3 => 'ADDIN' );

  my $sArch = 0;
  my ( $sCmd, $sCmdArgs, $sSwitch, $sHost, $sInstProfile,
       $iRetCode, $iInstId, $iTimeout ) = ();
  my $bNotOnlineFlag = 0;

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  map {
    $sArch |= 1 if /^jc$/;
    $sArch |= 2 if /^dw$/;
  } $sProcMon =~ /([^\s]+)/g;

  if ( $hsArch { $sArch } ) {
    $sSwitch = $hsArch { $sArch };
    &logTrace ( 20178,
    "<%s> Identified SAP Architecture [%s] from ProcMon [%s]",
    $sMyName, $sSwitch, $sProcMon );
  }
  else {
    &logTrace ( 20179,
    "<%s> Could not identify SAP Architecture from ProcMon [%s]",
    $sMyName, $sProcMon );
    $sSwitch = 'DEFAULT'; # Mark it default..
  }

  #----------------
  # Assign values..
  #----------------
  ( $sInstProfile = $sStartProfile ) =~ s/START_/${sSAPSID}_/;
  ( $sHost = $sStartProfile ) =~ s#.*/.*?_[^_]+_##; # Get hostname..
  ( $iInstId ) = $sInstName =~ /(\d{2})$/o;         # Get the last two..

  #-----------------------
  # Do the correct stuff..
  #-----------------------
  goto $sSwitch; {

    JAVA:

      if ( $sInstType =~ /^(dialog|central)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCmd = "$sSAPMonHome/jcmon";

        #-------- DO NOT INDENT --------#
        $sCmdArgs = "pf=$sInstProfile << EOF
20
0
0
EOF";
        #-------- DO NOT INDENT --------#

        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPNW04::ExecJCMon ( $sSAPAdmin,
                                            $sEnvFile,
                                            $sCmd,
                                            $sCmdArgs,
                                            $iTimeout );
      }
      else {
        &logError ( 20180,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      } 

    last;

    ABAP:

      if ( $sInstType =~ /^(dialog|central)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/dpmon";
        $sCmdArgs = "-info pf=$sInstProfile";
        $iTimeout = $iTillWhen - time;
        $iReturnCode = SAPNW04::ExecDpMon   ( $sSAPAdmin, # User context..
                                              $sEnvFile,  # Use envfile..
                                              $sCmd,
                                              $sCmdArgs,
                                              $iTimeout );

        #---------------------
        # If offline, return..
        #---------------------
        if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode  ||  
             VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
        elsif ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
          $bNotOnlineFlag = 1;
        }
        $sCmd = "$sSAPMonHome/sapinfo";
        $sCmdArgs = "ashost=$sHost sysnr=$iInstId";
        $iTimeout = $iTillWhen - time;
  
        $iReturnCode = SAPNW04::ExecSAPInfo ( $sSAPAdmin, # User context..
                                              $sEnvFile,  # Use envfile..
                                              $sCmd,
                                              $sCmdArgs,
                                              $iTimeout );
        if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
             VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
        elsif ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
          $bNotOnlineFlag = 1;
        }
        if( $sProcMon =~ /ms/ ){
          $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/msprot";
          $sCmdArgs = "-t -k 1 -n 2 -r 5 pf=$sInstProfile";
          $iTimeout = $iTillWhen - time;

          $iReturnCode = SAPNW04::ExecMsProt ( $sSAPAdmin,
                                               $sEnvFile,
                                               $sCmd,
                                               $sCmdArgs,
                                               $sInstType,
                                               $iTimeout );
	  if( VCSDef_MonitorExit::TIMEOUT ==  $iReturnCode ) {
             return $iReturnCode;
          }
        }
        if ( 1 == $bNotOnlineFlag && ( VCSDef_MonitorExit::OFFLINE != $iReturnCode ) ) {
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
        } 

      }
      else {
        &logError ( 20181,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

    last;

    ADDIN:

      if ( $sInstType =~ /^(dialog|central)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/dpmon";
        $sCmdArgs = "-info pf=$sInstProfile";
        $iTimeout = $iTillWhen - time;
        $iReturnCode = SAPNW04::ExecDpMon   ( $sSAPAdmin, # User context..
                                              $sEnvFile,  # Use envfile..
                                              $sCmd,
                                              $sCmdArgs,
                                              $iTimeout );

        #---------------------
        # If offline, return..
        #---------------------
        if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
             VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
        elsif ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
          $bNotOnlineFlag = 1;
        }

        $sCmd = "$sSAPMonHome/sapinfo";
        $sCmdArgs = "ashost=$sHost sysnr=$iInstId";
        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPNW04::ExecSAPInfo ( $sSAPAdmin, # User context..
                                              $sEnvFile,  # Use envfile..
                                              $sCmd,
                                              $sCmdArgs,
                                              $iTimeout );

        #---------------------
        # If offline, return..
        #---------------------
        if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
             VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
        elsif ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
          $bNotOnlineFlag = 1;
        }

        if( $sProcMon =~ /ms/ ){
          $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/msprot";
          $sCmdArgs = "-t -k 1 -n 2 -r 5 pf=$sInstProfile";
          $iTimeout = $iTillWhen - time;

          $iReturnCode = SAPNW04::ExecMsProt ( $sSAPAdmin,
                                               $sEnvFile,
                                               $sCmd,
                                               $sCmdArgs,
                                               $sInstType,
                                               $iTimeout );
          if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
               VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
            return $iReturnCode;
          }
          elsif ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
            $bNotOnlineFlag = 1;
          }
        }

        #--------------------------
        # Re-set command and args..
        #--------------------------
        if ( -x "/usr/sap/$sSAPSID/$sInstName/j2ee/os_libs/jcmon" ) { # SAPNW04..
          $sCmd = "/usr/sap/$sSAPSID/$sInstName/j2ee/os_libs/jcmon";
        }
        elsif ( -x "/usr/sap/$sSAPSID/SYS/exe/run/jcmon" ) { # SAPNW04s..
          $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/jcmon";
        }

        #-------- DO NOT INDENT --------#
        $sCmdArgs = "pf=$sInstProfile << EOF
20
0
0
EOF";
        #-------- DO NOT INDENT --------#

        $iTimeout = $iTillWhen - time;

        $iRetCode = SAPNW04::ExecJCMon ( $sSAPAdmin,
                                         $sEnvFile,
                                         $sCmd,
                                         $sCmdArgs,
                                         $iTimeout );
        #-------------------------------------------
        # Update, status only when it's not online..
        #-------------------------------------------
        unless ( VCSDef_MonitorExit::ONLINE == $iRetCode ) {
          $iReturnCode = $iRetCode;
        }
        if( VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
        if ( 1 == $bNotOnlineFlag && ( VCSDef_MonitorExit::OFFLINE != $iReturnCode ) ) {
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
        }
      }
      else {
        &logError ( 20182,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

    last;

    DEFAULT:

      if ( $sInstType !~ /^(dialog|central)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/ensmon";
        $sCmdArgs = "pf=$sInstProfile 2";
        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPNW04::ExecEnsMon ( $sSAPAdmin,
                                             $sEnvFile,
                                             $sCmd,
                                             $sCmdArgs,
                                             $sInstType,
                                             $iTimeout );
        #---------------------
        # If offline, return..
        #---------------------
        if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
             VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
        elsif ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
          $bNotOnlineFlag = 1;
        }
        if( $sProcMon =~ /ms/ ){
          $sCmd = "/usr/sap/$sSAPSID/SYS/exe/run/msprot";
          $sCmdArgs = "-t -k 1 -n 2 -r 5 pf=$sInstProfile";
          $iTimeout = $iTillWhen - time;

          $iReturnCode = SAPNW04::ExecMsProt ( $sSAPAdmin,
                                               $sEnvFile,
                                               $sCmd,
                                               $sCmdArgs,
                                               $sInstType,
                                               $iTimeout );
        }
        if( VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
        if ( 1 == $bNotOnlineFlag && ( VCSDef_MonitorExit::OFFLINE != $iReturnCode ) ) {
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
        }
      }
      else {
        &logError ( 20183,
        "<%s> Encountered unsupported InstType [%s]. Has to be one of
        [enqueue aenqueue jenqueue enqrep aenqrep jenqrep]",
        $sMyName, $sInstType );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

    last;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String SAPAdmin,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sSAPAdmin   UNIX user dedicated for this instance of SAP
 $sEnvFile    Full path to the environment file for SAP
 $sCmd        Absolute path to external monitor program
 $sCmdArgs    Command line arguments, if any, for Cmd
 $iTillWhen   Represents the upper limit set by engine in EPOCH 

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------
=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20184, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  TillWhen      [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdOut ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20185,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20186,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20187,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20188,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SAPNW04::ExecLocalCmd ( $sSAPAdmin,
                          $sEnvFile,
                          $sCmd,
                          $sCmdArgs,
                          \$iExitCode,
                          $iTimeout,
                          \@lsCmdOut );

  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
    &logError ( 20189,
    "<%s> Command [%s] with args [%s] timed out",
    $sMyName, $sCmd, $sCmdArgs );
    $iExitCode = 'undef';
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20190,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }
  else {
    &logWarn ( 20191,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference CmdOut );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdOut        Stores o/p from command if successful

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdOut ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20192, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdOut   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdOut );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdOut = (); # Reset..

  #-----------------------------------
  # Verify timeout is greater than 1..
  #-----------------------------------
  if ( $iTimeout <= 1 ) {
    &logError ( 20193,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    return 0;
  }

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20194,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdOut,
                                                           \$bTimedOut,
                                                           $iTimeout );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdOut;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20195,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20196,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdOut = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20197,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdOut [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20198,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20199, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20200,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  goto $sSwitch; {
    solaris:

      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;

    hpux:

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;

    aix:

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;


    linux:

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;
  }

  &logTrace ( 20201, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20202,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20203,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20204, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20205, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20206,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20207,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20208,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool DeletePidFile ( IN String ResourceName );


 Description : Deletes PID files for the VCS resource.

 Parameters :

 $sResourceName ResourceName for which the PID files needs to be deleted.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub DeletePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20319, "Subroutine <%s> called with:
  ResourceName        [%s]",

  $sMyName, $sResourceName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $sPidFileDir, $iRet ) = ();
  #-------------------
  # Delete PID files..
  #-------------------
  &logTrace ( 20320,
  "<%s> Deleting PID files for VCS resource [%s]",
  $sMyName, $sResourceName );

  $sPidFileDir = "/tmp/.VRTSSAPNW04/" . $sResourceName . "/PID" ;

  if ( -d $sPidFileDir ) {
    $iRet = system("/bin/rm -rf $sPidFileDir") ;

    if ( $iRet == 0 ) {
      &logTrace ( 20321,
      "<%s> Successfully deleted PID files in the directory [%s].",
      $sMyName, $sPidFileDir );
      $iReturnCode = 0;
    }
    else {
      &logWarn ( 20322,
      "<%s> Failed to delete PID files in directory [%s]. Remove the files manually.",
      $sMyName, $sPidFileDir );
      $iReturnCode = 1;
    }
  }
  else {
    &logTrace ( 20323,
    "<%s> PID files for VCS resource [%s] doesn't exist.",
    $sMyName, $sResourceName );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int ExecJCMon ( IN String  SAPAdmin,
                            IN String  EnvFile,
                            IN String  Cmd,
                            IN String  CmdArgs,
                            IN Integer Timeout );

 Description : Decides on status of SAP Instance based on jcmon command.

 Parameters :

 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sEnvFile            Full path to the environment file for SAP
 $sCmd                Full path to jcmon binary
 $sCmdArgs            Argument list for jcmon
 $iTimeout            Time in seconds for execution

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut


sub ExecJCMon {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20209, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, @lsCmdOut ) = ();

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20210,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20211,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20212,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( SAPNW04::ExecLocalCmd ( $sSAPAdmin, # User context..
                                  $sEnvFile,  # Use envfile..
                                  $sCmd,
                                  $sCmdArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20213,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );

    #------------------------------------------------------
    # Get status based on state of dispatcher and servers..
    #------------------------------------------------------
    $iReturnCode = SAPNW04::GetJCMonStatus ( \@lsCmdOut );
    @lsCmdOut = (); # Free..
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20214,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20215,
    "<%s> Cmd [%s] failed with exit code [%s]. Cmd output [%s]
    Returning [%s]", $sMyName, $sCmd, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int ExecMsProt ( IN String  SAPAdmin,
                             IN String  EnvFile,
                             IN String  Cmd,
                             IN String  CmdArgs,
                             IN String  InstType,
                             IN Integer Timeout );

 Description : Decides on status of SAP Instance based on msprot command.

 $sSAPAdmin   UNIX user dedicated for this instance of SAP
 $sEnvFile    Full path to the environment file for SAP
 $sCmd        Full path to msprot binary
 $sCmdArgs    Argument list for msprot 
 $sInstType   String identifier that classifies the SAP server instance type
 $iTimeout    Time in seconds for execution

 Parameters :

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExecMsProt {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $sInstType,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20324, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  InstType      [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $sInstType,
  $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sEntryPointName, $iExitCode, @lsCmdOut ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20325,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20326,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20327,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( SAPNW04::ExecLocalCmd ( $sSAPAdmin, # User context..
                                  $sEnvFile,  # Use envfile..
                                  $sCmd,
                                  $sCmdArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20328,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20329,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif ( 4 == $iExitCode ) {  # Server is down..
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;

    &logError ( 20330,
    "<%s> Cmd [%s] with args [%s] exited with [%s]. Cmd output [%s].
    Returning [%s]", $sMyName, $sCmd, $sCmdArgs, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }
  else { # Error..
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20331,
    "<%s> Cmd [%s] failed with exit code [%s]. Returning [%s]",
    $sMyName, $sCmd, $iExitCode, $iReturnCode );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int ExecEnsMon ( IN String SAPAdmin,
                             IN String EnvFile,
                             IN String Cmd,
                             IN String CmdArgs,
                             IN String InstType,
                             IN Integer Timeout );

 Description : Decides on status of SAP Instance based on ensmon command.

 $sSAPAdmin   UNIX user dedicated for this instance of SAP
 $sEnvFile    Full path to the environment file for SAP
 $sCmd        Full path to ensmon binary
 $sCmdArgs    Argument list for ensmon
 $sInstType   String identifier that classifies the SAP server instance type
 $iTimeout    Time in seconds for execution

 Parameters :

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExecEnsMon {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $sInstType,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20216, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  InstType      [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $sInstType,
  $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sEntryPointName, $iExitCode, @lsCmdOut ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20217,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20218,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20219,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( SAPNW04::ExecLocalCmd ( $sSAPAdmin, # User context..
                                  $sEnvFile,  # Use envfile..
                                  $sCmd,
                                  $sCmdArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20220,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20221,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif ( 4 == $iExitCode &&
          $sInstType =~ /^(|a|j)enqueue$/io ) {
    if ( $sEntryPointName eq 'monitor' ) {
      &logCritical ( 20222,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Replication Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
    else {
      &logTrace ( 20223,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Replication Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
  }
  elsif ( 8 == $iExitCode ) {  # Up already..
    if ( $sInstType =~ /^(|a|j)enqueue$/io ) { 
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;

      &logError ( 20224,
      "<%s> Cmd [%s] with args [%s] exited with [%s]. Returning [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, $iReturnCode );
    }
    elsif ( $sInstType =~ /^(|a|j)enqrep$/io ) {
      &logCritical ( 20225,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
  }
  else { # Error..
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20226,
    "<%s> Cmd [%s] failed with exit code [%s]. Returning [%s]",
    $sMyName, $sCmd, $iExitCode, $iReturnCode );
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int ExecDpMon   ( IN String SAPAdmin,
                              IN String EnvFile,
                              IN String Cmd,
                              IN String CmdArgs,
                              IN Integer Timeout );

 Description : Execute dpmon command in given timeout

 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sEnvFile            Full path to the environment file for SAP
 $sCmd                Full path to dpmon binary
 $sCmdArgs            Argument list for dpmon
 $iTimeout            Time in seconds for execution

 Parameters :

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExecDpMon {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20332, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, @lsCmdOut ) = ();

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20333,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20334,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20335,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( SAPNW04::ExecLocalCmd ( $sSAPAdmin, # User context..
                                  $sEnvFile,  # Use envfile..
                                  $sCmd,
                                  $sCmdArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20336,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20337,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif ( 156 == $iExitCode ) {  # Server is down..
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;

    &logError ( 20338,
    "<%s> Cmd [%s] with args [%s] exited with [%s]. Cmd output [%s].
    Returning [%s]", $sMyName, $sCmd, $sCmdArgs, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }
  else { # Error..
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20339,
    "<%s> Cmd [%s] failed with exit code [%s]. Cmd output [%s]
    Returning [%s]", $sMyName, $sCmd, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExecSAPInfo ( IN String SAPAdmin,
                              IN String EnvFile,
                              IN String Cmd,
                              IN String CmdArgs,
                              IN Integer Timeout );

 Description : Execute sapinfo command in given timeout

 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sEnvFile            Full path to the environment file for SAP
 $sCmd                Full path to sapinfo binary
 $sCmdArgs            Argument list for sapinfo
 $iTimeout            Time in seconds for execution

 Parameters :

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExecSAPInfo {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20227, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, @lsCmdOut ) = ();

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20228,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20229,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20230,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( SAPNW04::ExecLocalCmd ( $sSAPAdmin, # User context..
                                  $sEnvFile,  # Use envfile..
                                  $sCmd,
                                  $sCmdArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20231,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20232,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif ( 2 == $iExitCode ) {  # Server is down..
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;

    &logError ( 20233,
    "<%s> Cmd [%s] with args [%s] exited with [%s]. Cmd output [%s].
    Returning [%s]", $sMyName, $sCmd, $sCmdArgs, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }
  else { # Error..
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20234,
    "<%s> Cmd [%s] failed with exit code [%s]. Cmd output [%s]
    Returning [%s]", $sMyName, $sCmd, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetJCMonStatus ( IN RefStrList rlsCmdOut );

 Description : Decides on status of SAP Instance based on given jcmon
               output.

 Parameters :

 $rlsCmdOut      Reference to list containing output from jcmon execution.

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetJCMonStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rlsCmdOut ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20235, "Subroutine <%s> called with:
  rlsCmdOut         [%s]",
  $sMyName, $rlsCmdOut );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------------
  # Validate against empty list..
  #------------------------------
  return VCSDef_MonitorExit::UNKNOWN unless scalar @$rlsCmdOut;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sTmp, $sKey ) = ( '' ) x 2; 
  my ( $iSrv, $iStopped, $iRunning ) = ( 0 ) x 3; 
  my %hsCmdOut = map { /\b(dispatcher|server\d*)\b.*?\d\|(.*?)\s*\|/o } @$rlsCmdOut;

  foreach $sKey ( sort keys %hsCmdOut ) {
    $sTmp .= "Server Name [$sKey] State [$hsCmdOut{$sKey}]\n";
    if ( $sKey =~ /server/o ) {
      ++ $iSrv;
      ++ $iStopped if $hsCmdOut { $sKey } eq 'Stopped';
      ++ $iRunning if $hsCmdOut { $sKey } eq 'Running';
    }
  }

  &logTrace ( 20236, "<%s> SAP jcmon status : ---\n%s", $sMyName, $sTmp );

  #-----------------
  # Check on state..
  #-----------------
  if ( defined $hsCmdOut { dispatcher }      && # Defined..
       $hsCmdOut { dispatcher } eq 'Running' && # Dispatcher running..
       $iRunning ) {                            # One server running..
    &logTrace ( 20237,
    "<%s> Found Dispatcher and [%s] of [%s] server(s) running. Returning [%s]",
    $sMyName, $iRunning, $iSrv, $iReturnCode );
  }
  elsif ( defined $hsCmdOut { dispatcher }        && # Defined..
          ( $hsCmdOut { dispatcher } eq 'Stopped' || # Dispatcher stopped..
          $iStopped == $iSrv ) ) {                   # or all servers stopped..
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    &logError ( 20238,
    "<%s> Dispatcher or all [%s] server(s) stopped. Returning [%s]",
    $sMyName, $iSrv, $iReturnCode );
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logWarn ( 20239,
    "<%s> Status of SAP instance undecided:\n%sReturning [%s]",
    $sMyName, $sTmp, $iReturnCode );
  }

  ( $sTmp, $sKey ) = (); # Free..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool OnlineENQGrp ( IN String Sys,
                                IN String Grp,
                                IN String Reason,
                                IN String FaultedNode );

 Description : Pre-online trigger for SAPNW04

 Parameters :

 $sSys                System where the group is asked to be made online
 $sGrp                Group for which preonline has been invoked
 $sReason             Contains the string MANUAL or FAULT
 $sFaultedNode        Cluster node where the fault occurred due to FAULT

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineENQGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSys,
       $sGrp,
       $sReason,
       $sFaultedNode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20240, "Subroutine <%s> called with:
  Sys             [%s]
  Grp             [%s]
  Reason          [%s]
  FaultedNode     [%s]",

  $sMyName,
  $sSys, $sGrp, $sReason, $sFaultedNode );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..

  my %hsInstMatch = ( 'ENQREP'  => 'ENQUEUE',
                      'AENQREP' => 'AENQUEUE',
                      'JENQREP' => 'JENQUEUE' );

  my ( $i, $iExitCode, $sRes, $sDisp, $sGroup, $sNodeName,  $sEnqSrvRes,
       $sCmdArgs,   $sTempNode,    $sEnqRes,   $sInstType,   $sFailHost,
       $sKey,  $sSAPSID, @lsHaresOut, @lsCpy, @lsValidNodes, @lsEnqReps,
       @lsEnques,        %hsEnqueues,     %hsFltd,           %hsEnqReps,
       %hsEnqRepGrps, %hsEnqRepHosts, %hsSwEnqReps,    %hsOnlineEnqueues ) = ();

  my $sOSVer = ();

  #---------------------
  # Validate arguments..
  #---------------------
  if ( $sReason eq 'MANUAL' ) {
    &logTrace ( 20241,
    "<%s> Invoked during a normal switch-over or online.", $sMyName );
  }
  elsif ( $sSys =~ /\S/ && ! inet_aton ( $sSys ) ) {
    &logError ( 20242,
    "<%s> Destination cluster node [%s] cannot be resolved",
    $sMyName, $sSys );
    $iReturnCode = 0; # Error..
  }
  elsif ( $sFaultedNode ne 'UNDEF' && $sFaultedNode =~ /\S/ && ! inet_aton ( $sFaultedNode ) ) {
    &logError ( 20243,
    "<%s> Faulted cluster node [%s] cannot be resolved",
    $sMyName, $sFaultedNode );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -e $sCmd ) {
    &logError ( 20244,
    "<%s> Command [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20245,
    "<%s> Command [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------------------------------
  # Bug in VCS for FAULT. Verify group is really faulted..
  #-------------------------------------------------------
  # hasys -list SysState=RUNNING
  $sCmd = $sCmdHasys;
  $sCmdArgs = "-state";

  my ( @lsHasysRunning, @lsHasysFault, @lsHasysOut, %hsHasysRunning, %hsHasysFault ) = ();

  @lsHasysOut = readpipe ( "$sCmd $sCmdArgs" );
  shift @lsHasysOut;

  @lsHasysRunning = grep {
                  ( s/\n//  || 1 ) &&
                  ( s/.*:// || 1 ) &&   # for DR ...
                  ( s/^(\S+).*(RUNNING)$/$1/ )
                } @lsHasysOut;

  @lsHasysFault = grep {
                  ( s/\n//  || 1 ) &&
                  ( s/.*:// || 1 ) &&   # for DR ...
                  ( s/^(\S+).*(FAULTED|EXITED)$/$1/ )
                } @lsHasysOut;

  %hsHasysRunning = map { $_ => 1 } @lsHasysRunning;

  &logTrace ( 20246,
  "<%s> List of systems where VCS running [%s]",
  $sMyName, join ( ' ', @lsHasysRunning ) );

  %hsHasysFault   = map { $_ => 1 } @lsHasysFault;

  &logTrace ( 20247,
  "<%s> List of systems where VCS Faulted [%s]",
  $sMyName, join ( ' ', @lsHasysFault ) );

  $sCmd = $sCmdHares;
  $sCmdArgs = "-list Group=$sGrp State=FAULTED";

  @lsHaresOut = grep {
                  ( s/\n// || 1 ) &&
                  s/\s+.*//
                } readpipe ( "$sCmd $sCmdArgs" );

  #-------------------
  # Verify exit code..
  #-------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20248,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Group [%s] does not have any faulted resources.",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, $sGrp );

    #-----------------------------
    # Check for node FAULT event..
    #-----------------------------
    if ( $sFaultedNode eq 'UNDEF' && $sReason eq 'FAULT' ) {
      &logWarn ( 20249,
      "<%s> Missing faulted node name.
      Possible node FAULT event detected", $sMyName );
    }
    elsif ( $sFaultedNode eq 'UNDEF' && $sReason eq 'MANUAL' ) {
      &logWarn ( 20250,
      "<%s> Missing source node name.", $sMyName );
    }
    elsif ( ($sOSVer = $^O eq 'solaris' ? ( uname ) [2] : '') eq "5.10" ) {
      #---------------------------------------------------------------
      # Add zone specific support for Enqueue when zone itself faults.
      #---------------------------------------------------------------
      &logInfo ( 20251,
      "<%s> Checking if the Group [%s] is configured for Solaris non-global zones",
      $sMyName, $sGrp );

      my $sOut = ();
      my @lsOut = ();
      my $iCount = ();

      $sCmdArgs = "-display -attribute ContainerName -group $sGrp -type SAPNW04";
      @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCmd $sCmdArgs" );

      shift @lsHaresOut;

      foreach $sOut (@lsHaresOut) {
        @lsOut = split( '\s+', $sOut);
        $iCount = scalar @lsOut;
        if( (4 == $iCount) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
          &logWarn ( 20252,
          "<%s> ContainerName attribute is configured.
          Possible zone [%s] FAULT event detected", $sMyName, $lsOut[3] );

          last;
        }
      }
    }
  }
  else {
    &logTrace ( 20253,
    "<%s> Group [%s] has [%s] resource(s) in FAULTED state",
    $sMyName, $sGrp, join ( ' ', @lsHaresOut ) );
  }
                  
  #----------------
  # Re-initialise..
  #----------------
  $sCmdArgs = "-display -type SAPNW04 -attribute " .
              "State Group InstType EnqSrvResName SAPSID";

  #---------------------------------------------
  # Assume the group name is valid and proceed..
  #---------------------------------------------
  @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sCmd $sCmdArgs" );

  #-----------------------------------------
  # Check if command executed successfully..
  #-----------------------------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20254,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    return 0; # Error..
  }
  elsif ( $lsHaresOut[0] =~ /10554.*No resource exists with/o ) {
    &logError ( 20255,
    "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
    $sMyName, $sCmd, $sCmdArgs,  $lsHaresOut[0] );
    return 0; # Error..
  }

  $sFailHost = $sSys;    # Set to default..
  push @lsHaresOut, "#"; # Insert trailing de-marker..
  @lsCpy = @lsHaresOut;  # Backup..

  #---------------------------------------------------
  # Mark all ENQUEUE/ENQREP resources in given group..
  #---------------------------------------------------
  @lsEnques =
  grep {
     ( s/\n// || 1 )                   && # Chomp, and always make true..
     ( /Group.*\s+(\S+)/               && # Mark the group name..
     ( $sGroup = $1 ) && 0 )           || # Store, and always make false..
     ( s/^(\S+).*InstType.*\s             # Mark resource..
     ((|A|J)ENQ(UEUE|REP))/$1:$2/ix    && # If ENQUEUE or ENQREP variant..
     ++ $i )                           || # Set flag..
     ( $i                              && # If flag..
     s/.*SAPSID.*\s+// )               || # Get the SAPSID..
     ( $i                              && # If flag..
     s/#/sprintf ( $sGroup, -- $i )/e )   # Print group.. reset flag..
  } @lsCpy; 

  &logTrace ( 20256,
  "<%s> Searching ENQUEUE resource in group [%s]", $sMyName, $sGrp );

  &logTrace ( 20257,
  "<%s> Enques values are [%s]", $sMyName, join(' ', @lsEnques) );

  #-------------------
  # Start processing..
  #-------------------
  while ( 3 == ( ( $sRes, $sSAPSID, $sGroup ) = splice ( @lsEnques, 0, 3 ) ) ) {
    ( $sRes, $sInstType ) = $sRes =~ /([^:]+)/g;

    #-------------------------------
    # Passed group has ENQ(UE|REP)..
    #-------------------------------
    if ( $sGrp eq $sGroup ) {
      if ( $sInstType =~ /ENQUEUE/io ) {
        &logInfo ( 20258,
        "<%s> +------- Group [%s] has ENQUEUE variant -------+",
        $sMyName, $sGrp );

        #--------
        # Store..
        #--------
        push @{ $hsEnqueues { $sRes } },
        $sSAPSID, $sInstType, $sGroup; # Store..
      }
      elsif ( $sInstType =~ /ENQREP/io ) {
        &logInfo ( 20259,
        "<%s> +------- Group [%s] has ENQREP variant -------+",
        $sMyName, $sGrp );
        
        #--------
        # Store..
        #--------
        push @{ $hsSwEnqReps { $sRes } },
        $sSAPSID, $sInstType, $sGroup; # Store..
      }
    }

    &logTrace ( 20260,
    "<%s> Resource [%s] SAPSID [%s] InstType [%s] found in Group [%s]",
    $sMyName, $sRes, $sSAPSID, $sInstType, $sGroup );
  }

  #-----------------------------------------------
  # Find online resources and list their details..
  #-----------------------------------------------
  @lsCpy = @lsHaresOut; # Re-initialise..
  $i = 0;
  
  &logTrace ( 20261,
  "<%s> Looking for all ONLINE SAPNW04 resources", $sMyName );
  
  @lsEnqReps =
  grep {
    ( s/\n// || 1 )                        && # Chomp, and always make true..
    ( /^(\S+).*Group.*\s+(\S+)/            && # Mark resource, and group name..
    ( $sGroup = "$1:$2" ) && 0 )           || # Store, and always make false..
    ( s/.*\s(\S+)\s+ONLINE$/$1/ && ++ $i ) || # If online, get node name, set flag..
    ( $i                                   && # If flag..
    /EnqSrvResName|InstType|SAPSID/        && # and contains one of these..
    s/.*\s+// )                            || # get the last field..
    ( $i                                   && # If flag, use demarker..
    s/#/sprintf ( $sGroup, -- $i )/e )        # to print the group name.. reset flag..
  } @lsCpy;                                   # From the copy..
  
  &logTrace ( 20262,
  "<%s> EnqReps [%s]", $sMyName, join(' ', @lsEnqReps) );

  while ( 5 == ( ( $sNodeName, $sEnqSrvRes, $sInstType, $sSAPSID, $sGroup ) = splice ( @lsEnqReps, 0, 5 ) ) ) {

    $sNodeName =~ s/.*://; # Remove cluster component, under GCO..
    ( $sRes, $sGroup ) = $sGroup =~ /([^:]+)/g;

    #----------------------------------
    # Store only ENQ(UE|REP) variants..
    #----------------------------------
    if ( $sInstType =~ /ENQREP/io ) {
      &logTrace ( 20263, "<%s> Located ENQREP variant. Resource [%s] details:
      Group [%s] Node [%s] EnqSrvResName [%s] InstType [%s] SAPSID [%s]",
      $sMyName, $sRes, $sGroup, $sNodeName, $sEnqSrvRes, $sInstType, $sSAPSID );
 
      push @{ $hsEnqReps { $sRes } },
      $sGroup, $sEnqSrvRes, $sInstType, $sSAPSID, $sNodeName;
    }
    elsif ( $sInstType =~ /ENQUEUE/io ) {
      $sEnqSrvRes = ""; # Truncate..
      &logTrace ( 20264, "<%s> Located ENQUEUE variant. Resource [%s] details:
      Group [%s] Node [%s] EnqSrvRes [%s] InstType [%s] SAPSID [%s]",
      $sMyName, $sRes, $sGroup, $sNodeName, $sEnqSrvRes, $sInstType, $sSAPSID );
 
      push @{ $hsOnlineEnqueues { $sRes } },
      $sGroup, $sEnqSrvRes, $sInstType, $sSAPSID, $sNodeName;
    }
  }
  
  #--------------------------------------------
  # Validate if group has ENQ(UE|REP) variant..
  #--------------------------------------------
  @lsCpy = @lsHaresOut; # Re-initialise.

  if ( keys %hsEnqueues ) {
    &logInfo ( 20265,
    "<%s> Failover Group [%s] has ENQUEUE variant resource(s) [%s].",
    $sMyName, $sGrp, join ( ' ' , keys %hsEnqueues ) );
  }
  elsif ( keys %hsSwEnqReps ) {
    #------------------------------------------
    # Get all valid nodes for the faulted REP..
    #------------------------------------------
    @lsCpy = @lsHaresOut;          # Re-initialise.
    $sRes =  join ( '|', keys %hsSwEnqReps );
    $sDisp =  join ( ' ', keys %hsSwEnqReps );

    &logInfo ( 20266,
    "<%s> Failover Group [%s] has ENQREP variant resource(s) [%s]",
    $sMyName, $sGrp, $sDisp );

    %hsFltd = map { $_ => 1 }                 # All faulted nodes..
    grep {                                    # Grep'ed here..
      ( s/\n// || 1 )                    &&   # Chomp, and always make true..
      s/^($sRes).*State\s+(\S+).*FAULTED/$2/  # Pick all nodes where FAULTed..
    } @lsCpy;                                 # From cmd output..

    @lsCpy = @lsHaresOut;                     # Re-initialise.
    @lsValidNodes = 
    grep {
      ( s/\n// || 1 )                       &&  # Chomp, and always make true..
      s/^($sRes).*State\s+(\S+).*LINE/$2/   ||  # Pick the node not FAULTed..
      s/^($sRes).*EnqSrvResName.*\s+(\S+)/::$2/ # And EnqSrvResName..
    } @lsCpy;                                   # From cmd output..

    #-------------------------------------------------
    # Now extract EnqSrvResName and configured nodes..
    #-------------------------------------------------
    @lsCpy = @lsValidNodes; # Re-use..
    @lsValidNodes = ();     # Truncate..

    foreach $sKey ( @lsCpy ) {
      if ( $sKey =~ /^::/ ) { # EnqResName found..
        ( $sEnqRes = $sKey ) =~ s/:://;
      }
      else {
        $sKey =~ s/.*://; # Remove cluster component, under GCO..
        push ( @lsValidNodes, $sKey ) unless $hsFltd { $sKey };
      }
    }

    &logTrace ( 20267,
    "<%s> Nodes on which Service Group not in FALUTED state [%s]",
    $sMyName, join(' ', @lsValidNodes) );

    ( $sRes ) = keys %hsSwEnqReps; # Take one..

    #--------------------
    # Remove duplicates..
    #--------------------
    %hsFltd = (); # Truncate..
    @lsValidNodes = grep { ! $hsFltd { $_ } ++ } @lsValidNodes;

    &logTrace ( 20268,
    "<%s> One of ENQREP resource(s) [%s] defines EnqSrvResName as [%s].

    Details for ENQREP resource [%s]:

    InstType [%s] SAPSID [%s]. Possible failover nodes [%s]",
    $sMyName, $sDisp, $sEnqRes, $sRes,
    ${ $hsSwEnqReps { $sRes } } [1], ${ $hsSwEnqReps { $sRes } } [0],
    join ( ' ', @lsValidNodes ) );

    #--------------------------------------------------------
    # Check if the node on which service group is configured
    # is Running at this moment..
    #--------------------------------------------------------
    @lsValidNodes = grep { $hsHasysRunning { $_ } } @lsValidNodes;

    &logTrace ( 20269,
    "<%s> Valid nodes for ENQREP failover targets [%s]",
    $sMyName, join ( ' ', @lsValidNodes ) );

    #--------------------------
    # If online ENQUEUE found..
    #--------------------------
    if ( defined $hsOnlineEnqueues { $sEnqRes } ) {
      &logTrace ( 20270,
      "<%s> Details for defined ENQUEUE resource [%s]:
      InstType [%s] SAPSID [%s] online on node [%s]",
      $sMyName, $sEnqRes, ${ $hsOnlineEnqueues { $sEnqRes } } [2],
      ${ $hsOnlineEnqueues { $sEnqRes } } [-2],
      ${ $hsOnlineEnqueues { $sEnqRes } } [-1] );
    }
    else {
      &logWarn ( 20271,
      "<%s> Corresponding ENQUEUE resource(s) is NOT online in the cluster", $sMyName );
    }

    #-----------------------------
    # Now validate EnqSrvResName..
    #-----------------------------
    if ( defined $hsOnlineEnqueues { $sEnqRes }             && # Defined..
         ( $hsInstMatch { ${ $hsSwEnqReps { $sRes } } [1] } eq # InstType pair..
         ${ $hsOnlineEnqueues { $sEnqRes } } [2] )          && # is a match..
         ( ${ $hsSwEnqReps { $sRes } } [0]                  eq # SAPSID..
           ${ $hsOnlineEnqueues { $sEnqRes } } [-2] ) ) {      # matches..

      $sTempNode = ${ $hsOnlineEnqueues { $sEnqRes } } [-1];   # Get the online node..

      ( $sFailHost ) = grep {                                  # Find the node other than..
                         ! /^($sTempNode|$sFaultedNode)$/o &&  # ENQUEUE online or faulted..
                         ! $hsHasysFault{$_}                   # nodes faulted..
                       } @lsValidNodes;                        # From all posibilities..

      #---------------------------------------------------
      # Check if we are in a position to failover safely..
      #---------------------------------------------------
      if ( $sFaultedNode ne 'UNDEF' || $sReason ne 'MANUAL' ||
           ( $sFaultedNode eq 'UNDEF' && $sReason eq 'MANUAL' && (scalar @lsHasysFault || $sTempNode eq $sSys)) ) {
        if ( defined $sFailHost ) {
          &logInfo ( 20272,
          "<%s> Rejecting ENQUEUE node [%s]. Failing over group [%s] to node [%s]",
          $sMyName, $sTempNode, $sGrp, $sFailHost );
        }
        else { # Prevent onlining..
          &logError ( 20273,
          "<%s> Could not locate any node to online Group [%s]",
          $sMyName, $sGrp );
          return $iReturnCode; # Error.. Prevent online of ENQREP..
        }
      }

      #-------------------------------
      # Online group to default node..
      #-------------------------------
      if ( SAPNW04::OnlineGrp ( $sGrp,
                                $sFailHost ) ) {
        &logInfo ( 20274,
        "<%s> Issued online of Group [%s] on node [%s]",
        $sMyName, $sGrp, $sFailHost );
        return $iReturnCode; # Success..
      }
      else {
        &logError ( 20275,
        "<%s> Could not issue online of Group [%s] on node [%s]",
        $sMyName, $sGrp, $sFailHost );
        return 0; # Error..
      }
    }
    else {
      &logError ( 20276,
      "<%s> Failover group [%s] does not have any online ENQ(UEUE|REP) variant",
      $sMyName, $sGrp );
      return 0; # Do nothing..
    }
  }
  else {
    &logInfo ( 20277,
    "<%s> Failover group [%s] does not have any ENQ(UEUE|REP) variant",
    $sMyName, $sGrp );
    return 0; # Do nothing..
  }
  
  #------------------------------------------------
  # Validate if we have any online ENQREP variant..
  #------------------------------------------------
  &logTrace ( 20278,
  "<%s> +----- ENQREP online node and group details -----+", $sMyName );

  #----------------------------------------------
  # Remove ENQREPs whose EnqRepSrv aren't found..
  #----------------------------------------------
  foreach $sKey ( keys %hsEnqReps ) {
    if ( ! $hsEnqueues { ${$hsEnqReps { $sKey } }[1] } ) { # Not in passed group..
      &logTrace ( 20279,
      "<%s> EnqResName [%s] resource not found in Grp [%s].
      Marked resource [%s] invalid",
      $sMyName, ${$hsEnqReps { $sKey } }[1], $sGrp, $sKey );

      delete $hsEnqReps { $sKey };                        # Then delete..
    }
    else {
      ++ $hsEnqRepHosts { ${$hsEnqReps { $sKey } }[-1] }; # Else store online node..
      ++ $hsEnqRepGrps  { ${$hsEnqReps { $sKey } }[0] };  # Else store group..

      &logTrace ( 20280,
      "<%s> Online node [%s] Group [%s]",
      $sMyName, ${$hsEnqReps{$sKey}}[-1], ${$hsEnqReps{$sKey}}[0] );
    }
  }

  #------------------------------------------------
  # Validate if we have any online ENQREP variant..
  #------------------------------------------------
  if ( ! keys %hsEnqReps && $sReason ne 'MANUAL' ) {
    $sFailHost = $sSys; # Fail-over to the default node..
    &logCritical ( 20281,
    "<%s> Corresponding ENQREP server(s) not found ONLINE in cluster.
    Onlining group [%s] on node [%s]", $sMyName, $sGrp, $sFailHost );
  
    #-------------------------------
    # Online group to default node..
    #-------------------------------
    if ( SAPNW04::OnlineGrp ( $sGrp,
                              $sFailHost ) ) {
      &logInfo ( 20282,
      "<%s> Issued online of Group [%s] on node [%s]",
      $sMyName, $sGrp, $sFailHost );
      return $iReturnCode; # Success..
    }
    else {
      &logError ( 20283,
      "<%s> Could not issue online of Group [%s] on node [%s]",
      $sMyName, $sGrp, $sFailHost );
      return 0; # Error..
    }
  }
  
  #-----------------------------------------------
  # Check if we have ENQREPs in different groups..
  #-----------------------------------------------
  if ( keys %hsEnqRepGrps > 1 ) {
    &logCritical ( 20284,
    "<%s> ENQREP configured in different service groups [%s]",
    $sMyName, join ( ' ', keys %hsEnqRepGrps ) );
  }
  
  #----------------------------------------------
  # Now check if we have a unique failover node..
  #----------------------------------------------
  if ( keys %hsEnqRepHosts > 1 ) {
    &logCritical ( 20285,
    "<%s> ENQREP found online on different nodes [%s]",
    $sMyName, join ( ' ', keys %hsEnqRepHosts ) );
  
    #------------------------------------------
    # Delete entries for all other than JENQREP..
    #------------------------------------------
    foreach $sKey ( keys %hsEnqReps ) {
      if ( ${$hsEnqReps { $sKey } }[2] ne 'JENQREP' ) {
        &logError ( 20286,
        "<%s> Marking resource [%s] with InstType [%s] as invalid",
        $sMyName, $sKey, ${$hsEnqReps { $sKey } }[2] );
  
        #--------------------------
        # Delete in correct order..
        #--------------------------
        delete $hsEnqRepHosts { ${$hsEnqReps { $sKey } }[-1] }; # Delete node first..
        delete $hsEnqRepGrps  { ${$hsEnqReps { $sKey } }[0] };  # Delete group next..
        delete $hsEnqReps { $sKey };                            # Now the resource..
      }
    }
  }

  #-------------------------
  # Find out failover node..
  #-------------------------
  foreach $sKey ( keys %hsEnqReps ) {
    &logTrace ( 20287,
    "<%s> EnqRep server [%s] details:
    Group [%s] EnqSrvRes [%s] InstType [%s] SAPSID [%s] online on [%s]",
    $sMyName, $sKey,
    ${ $hsEnqReps { $sKey } }[0], ${ $hsEnqReps { $sKey } }[1],
    ${ $hsEnqReps { $sKey } }[2], ${ $hsEnqReps { $sKey } }[3],
    ${ $hsEnqReps { $sKey } }[4] );
  
    #-----------------------------------------------------------
    # Enqueue has been defined in EnqResName of ENQREP variant..
    #-----------------------------------------------------------
    if ( scalar @{ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } } ) { # EnqResName is correct..
      if ( ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0] ne    # SAPSID of ENQUEUE variant..
                            ${$hsEnqReps { $sKey }}[3] ) {          # ne SAPSID of ENQREP..
        &logError ( 20288,
        "<%s> SAPSID [%s] of ENQREP resource [%s] differs from
        SAPSID [%s] of ENQUEUE resource [%s]",
        $sMyName, ${$hsEnqReps { $sKey }}[3], $sKey,
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0],
        ${ $hsEnqReps { $sKey } }[1] );
      }
      elsif ( uc ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1] ne # ENQUEUE variant..
              $hsInstMatch { uc ${$hsEnqReps {$sKey }}[2] } ) {        # not correct match..
        &logError ( 20289,
        "<%s> InstType [%s] of ENQREP resource [%s] not a match for
        InstType [%s] of ENQUEUE resource [%s]", $sMyName,
        ${$hsEnqReps { $sKey }}[2], $sKey,
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1],
        ${ $hsEnqReps { $sKey } }[1] );
      }
      else { # All okay..
        &logTrace ( 20290,
        "<%s> EnqSrvResName [%s] matches with:
        ENQUEUE SAPSID [%s] InsType [%s] Group [%s]",
        $sMyName, ${$hsEnqReps { $sKey }}[1],
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0],
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1],
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[2] );
  
        #-----------------------------------------------------
        # Failover to ENQREP's node, if not the FAULTED node..
        #-----------------------------------------------------
        $sFailHost = ${ $hsEnqReps { $sKey } }[-1] eq $sFaultedNode ? $sSys :
                     ${ $hsEnqReps { $sKey } }[-1];
      }
    }
    else { # EnqResName is wrong..
      &logError ( 20291,
      "<%s> EnqResName [%s] value is in-correct for ENQREP resource [%s]",
      $sMyName,  ${$hsEnqReps { $sKey }}[1], $sKey );
    }
  }
  
  &logInfo ( 20292, "<%s> Failing over Group [%s] to node [%s]",
  $sMyName, $sGrp, $sFailHost );

  #-------------------------------
  # Online group to default node..
  #-------------------------------
  if ( SAPNW04::OnlineGrp ( $sGrp,
                            $sFailHost ) ) {
    &logInfo ( 20293, "<%s> Issued online of Group [%s] on node [%s]",
    $sMyName, $sGrp, $sFailHost );
  }
  else {
    &logError ( 20294,
    "<%s> Could not issue online of Group [%s] on node [%s]",
    $sMyName, $sGrp, $sFailHost );
    $iReturnCode = 0; # Do nothing..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool OnlineGrp ( IN String Grp,
                             IN String Sys );

 Description : Pre-online trigger for SAPNW04

 Parameters :

 $Grp                 Cluster group to online
 $sSys                System where the group is asked to be made online

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sGrp,
       $sSys ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20295, "Subroutine <%s> called with:
  Grp             [%s]
  Sys             [%s]",

  $sMyName, $sGrp, $sSys );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHagrp; # Get global..
  my @lsCmdArgs = ( "-online", "-nopre", "$sGrp", "-sys", "$sSys" );
  my $iExitCode = undef;

  &logTrace ( 20296, "<%s> Executing Command [%s] with arguments [%s]",
  $sMyName, $sCmd, join( ' ', @lsCmdArgs ) );

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ! -e $sCmd ) {
    &logError ( 20297,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20298,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  else { # Execute..
    &logTrace ( 20299,
    "<%s> Executing Cmd [%s] with args [%s]",
    $sMyName, $sCmd, join ( ' ', @lsCmdArgs ) );

    system ( $sCmd, @lsCmdArgs ); # Execute..

    #-------------------------
    # Check execution status..
    #-------------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20300,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20301,
      "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
    }
  }

  return $iReturnCode;
}

1; # So PERL knows its included..
